//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002,2008 Oracle.  All rights reserved.
//
// $Id: DictionaryCache.cpp,v 1.7 2008/02/20 17:20:17 gmf Exp $
//

#include "DictionaryCache.hpp"
#include "DictionaryDatabase.hpp"
#include "ScopedDbt.hpp"

using namespace DbXml;

int DictionaryCache::hashSize_ = 79;

namespace DbXml
{

//
// This implementation relies on the fact that the values never
// disappear -- the pointers must remain valid
//	
class DictionaryCacheEntry {
public:
	DictionaryCacheEntry(nameId_t nid, void *value,
			     int valueLen) : nid_(nid), next_(0), len_(valueLen)
	{
		// allocator created enough space for copy
		memcpy(((char*)this) + sizeof(*this), value, valueLen);
	}
	DictionaryCacheEntry *getNext() const {
		return next_;
	}
	void setNext(DictionaryCacheEntry *next) {
		next_ = next;
	}
	nameId_t getNid() const {
		return nid_;
	}
	int getLen() const {
		return len_;
	}
	// value starts just after nid and next
	const char *getValue() const {
		return ((char *)this) + sizeof(*this);
	}
private:
	nameId_t nid_;
	DictionaryCacheEntry *next_;
	int len_;
	// value follows
};

class DictionaryCacheBuffer {
public:
	DictionaryCacheBuffer(bool isFirst = false) : current_(0), next_(0),
		buffer_(0), capacity_(dcacheBufferSize_), used_(0)
	{
		if (isFirst)
			current_ = this;
		buffer_ = (char *) ::malloc(dcacheBufferSize_);
		if (!buffer_)
			throw XmlException(
				XmlException::NO_MEMORY_ERROR,
				"Failed to allocate memory for DictionaryCache");
	}
	
	~DictionaryCacheBuffer() {
		// semi-recursive.  If this is ever a problem, change it
		if (next_) {
			delete next_;
			next_ = 0;
		}
		::free(buffer_);
	}
	// not in-line -- it does a bit of work
	DictionaryCacheEntry *allocateEntry(int len, dbxml_mutex_t mutex);
private:
	// 4-byte alignment
	static int align(int sz) {
		return ((sz + 3) & ~3);
	}
	DictionaryCacheBuffer *current_;
	DictionaryCacheBuffer *next_;
	char *buffer_;
	int capacity_;
	int used_;
	static const int dcacheBufferSize_;
};
}

const int DictionaryCacheBuffer::dcacheBufferSize_ = 4096;

DictionaryCache::DictionaryCache()
	: ddb_(0), htable_(0), mem_(0), mutex_(MutexLock::createMutex())
{
	htable_ = new DictionaryCacheEntry*[hashSize_];
	if (htable_) {
		memset(htable_, 0, hashSize_ * sizeof(DictionaryCacheEntry*));
		mem_ = new DictionaryCacheBuffer(true);
		if (mem_) {
			return;
		}
	}
	throw XmlException(XmlException::NO_MEMORY_ERROR,
			   "Failed to allocate memory for DictionaryCache");
}

DictionaryCache::~DictionaryCache()
{
	if (mem_)
		delete mem_;
	if (htable_)
		delete [] htable_;
	MutexLock::destroyMutex(mutex_);
}


// Basic lookup, returning a pointer that will be valid
// until this object is deleted.  Lookup does not require locking
// because entries are never removed, and they are only added
// at the end of a hash chain with a simple pointer update.
const char *
DictionaryCache::lookup(OperationContext &context, const NameID &id,
			bool useDictionary)
{
	while (true) { // will only ever loop once
		nameId_t nid = id.raw();
		int bucket = hash(nid);
		DictionaryCacheEntry *current = htable_[bucket];
		while (current && (current->getNid() != nid))
			current = current->getNext();
		if (current)
			return current->getValue();
		if (!useDictionary)
			return 0;
		// read through the dictionary
		const char *name = 0;
		int ret = ddb_->lookupStringNameFromID(context, id,
						       &name);
		if (ret == 0) {
			DBXML_ASSERT(name);
			insert(nid, name);
		} else {
			// this should never happen, but if it does,
			// be silent (for now)
			return 0;
		}
	}
}

// Basic lookup, returning a pointer that will be valid
// until this object is deleted.  Lookup does not require locking
// because entries are never removed, and they are only added
// at the end of a hash chain with a simple pointer update.
bool
DictionaryCache::lookup(OperationContext &context, const NameID &id,
			DbtOut &dbt, bool useDictionary)
{
	while (true) { // will only ever loop once
		nameId_t nid = id.raw();
		int bucket = hash(nid);
		DictionaryCacheEntry *current = htable_[bucket];
		while (current && (current->getNid() != nid))
			current = current->getNext();
		if (current) {
			dbt.set(current->getValue(), current->getLen());
			return true;
		}
		if (!useDictionary)
			return false;
		// read through the dictionary
		int ret = ddb_->lookupStringNameFromID(context, id, dbt);
		if (ret == 0) {
			DBXML_ASSERT(dbt.get_size());
			insert(nid, dbt);
		} else {
			// this should never happen, but if it does,
			// be silent (for now)
			return false;
		}
	}
}

// inserting at the end of a chain makes it safe to not lock
// on lookup
void
DictionaryCache::insert(const nameId_t nid, const char *value)
{
	DbtIn val((void *)value, ::strlen(value) + 1);
	insert(nid, val);
}

// inserting at the end of a chain makes it safe to not lock
// on lookup
void
DictionaryCache::insert(const nameId_t nid, const Dbt &value)
{
	// will throw if no memory
	DictionaryCacheEntry *newEntry = mem_->allocateEntry(
		value.get_size(), mutex_);
	// placement new
	(void) new (newEntry) DictionaryCacheEntry(nid, value.get_data(),
						   value.get_size());
	int bucket = hash(nid);
	MutexLock ml(mutex_);
	DictionaryCacheEntry *current = htable_[bucket];
	if (!current) {
		htable_[bucket] = newEntry;
	} else {
		// check for duplicate and find last entry...
		DictionaryCacheEntry *last = 0;
		while (current && (current->getNid() != nid)) {
			last = current;
			current = current->getNext();
		}
		if (!current) {
			DBXML_ASSERT(last);
			last->setNext(newEntry);
		}
	}
}

DictionaryCacheEntry *
DictionaryCacheBuffer::allocateEntry(int valueLen, dbxml_mutex_t mutex) {
	int needed = valueLen + sizeof(DictionaryCacheEntry);
	needed = align(needed); // round up to 4-byte boundary

	MutexLock ml(mutex);
	if ((current_->used_ + needed) > current_->capacity_) {
		ml.unlock(); // unlock for allocation; note: this can race
		DictionaryCacheBuffer *newbuf =
			new DictionaryCacheBuffer();
		if (!newbuf)
			throw XmlException(
				XmlException::NO_MEMORY_ERROR,
				"Failed to allocate memory for DictionaryCache");
		ml.lock(); // re-lock, and check state again
		if ((current_->used_ + needed) > current_->capacity_) {
			current_->next_ = newbuf;
			current_ = newbuf;
		} else
			delete newbuf; // lost the race
	}
	DBXML_ASSERT((current_->used_ % 4) == 0);
	DictionaryCacheEntry *retVal = (DictionaryCacheEntry *)
		(current_->buffer_+current_->used_);
	current_->used_ += needed;
	return retVal;
}
