//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002,2008 Oracle.  All rights reserved.
//
// $Id: Manager.hpp,v 1.41 2008/03/26 19:35:19 gmf Exp $
//

#ifndef __MANAGER_HPP
#define __MANAGER_HPP

#include <string>
#include <map>

#include "ReferenceCounted.hpp"
#include "Container.hpp"
#include "Globals.hpp"
#include "Log.hpp"
#include "DocID.hpp"

class DbEnv;

namespace DbXml
{

class Container;
class ContainerBase;
class Transaction;
class ResolverStore;
class DictionaryDatabase;
class RecordDatabase;
class DocDatabase;
class Document;
class QueryPlanRoot;
	
class Manager : public Log, public ReferenceCountedProtected
{
public:
	/**
	 * It's assumed that the DbEnv is already open. If adopted, the DbEnv
	 * will be both closed and deleted by the XmlDocument when it is
	 * destroyed.
	 */
	Manager(DbEnv *dbEnv, u_int32_t flags = 0);
	Manager(u_int32_t flags);
	~Manager();


	//
	// Timezone methods
	//

	/// Timezone is measured as an offset in seconds
	int getImplicitTimezone() const;

	/// Timezone is measured as an offset in seconds
	void setImplicitTimezone(int tz);

	//
	// Resolver methods
	//

	ResolverStore &getResolverStore() {
		return *resolvers_;
	}

	u_int32_t getFlags() const {
		return flags_;
	}
	
	//
	// Container
	//

	void setDefaultContainerFlags(u_int32_t flags) {
		lock(); defaultContainerFlags_ = flags; unlock();
	}
	u_int32_t getDefaultContainerFlags() { return defaultContainerFlags_; }

	void setDefaultPageSize(u_int32_t pageSize) {
		lock(); defaultPageSize_ = pageSize; unlock();
	}
	u_int32_t getDefaultPageSize() { return defaultPageSize_; }

	void setDefaultSequenceIncrement(u_int32_t incr) {
		lock(); defaultSequenceIncr_ = incr; unlock();
	}
	u_int32_t getDefaultSequenceIncrement() { return defaultSequenceIncr_; }

	XmlContainer::ContainerType getDefaultContainerType() {
		return defaultContainerType_;
	}
	void setDefaultContainerType(XmlContainer::ContainerType type) {
		lock(); defaultContainerType_ = type; unlock();
	}

	XmlContainer openContainer(const std::string &name,
				   Transaction *txn, u_int32_t flags,
				   XmlContainer::ContainerType type,
				   int mode,
				   bool doVersionCheck = true);
	XmlContainer getOpenContainer(const std::string &name);

	// factory methods for XmlManager
	Document *createDocument();
	Transaction *createTransaction(DbTxn *toUse);
	Transaction *createTransaction(u_int32_t flags = 0);
	XmlInputStream *createLocalFileInputStream(
		const std::string &filename) const;
	XmlInputStream *createMemBufInputStream(
		const char *srcDocBytes,
		const unsigned int byteCount,
		const char *const bufId,
		bool adoptBuffer = false) const;

	XmlInputStream *createMemBufInputStream(
		const char *srcDocBytes,
		const unsigned int byteCount,
		bool copyBuffer) const;
	XmlInputStream *createURLInputStream(const std::string &baseId, 
					    const std::string &systemId, 
					    const std::string &publicId) const;
	XmlInputStream *createURLInputStream(const std::string &baseId, 
					    const std::string &systemId) const;
	

	void removeContainer(Transaction *txn, const std::string &name);
	void renameContainer(Transaction *txn, const std::string &oldName,
			     const std::string &newName);

	QueryPlanRoot *getDefaultQueryPlanRoot() const { return defaultQPR_; }
	ContainerBase *getContainerFromID(int id, bool acquire) const;
	
	DbEnv *getDbEnv() { return dbEnv_; }

	const std::string &getHome() const { return dbHome_; }

	bool isTransactedEnv() const { return ((dbEnvOpenFlags_ & DB_INIT_TXN) != 0); }
	bool isLockingEnv() const { return ((dbEnvOpenFlags_ & DB_INIT_LOCK) != 0); }
	bool isCDBEnv() const { return ((dbEnvOpenFlags_ & DB_INIT_CDB) != 0); }
	bool allowAutoOpen() const { return autoOpen_; }

	//
	// Logging
	//
	virtual void log(ImplLogCategory c, ImplLogLevel l,
			 const std::ostringstream &s) const;
	virtual void log(ImplLogCategory c, ImplLogLevel l,
			 const std::string &s) const;
	virtual void checkFlags(const FlagInfo *flag_info,
				const char *function,
				u_int32_t flags, u_int32_t mask) const;

	// create/get in-memory dictionary for handling
	// transient document parsing/querying.  Once created,
	// this remains open for the lifetime of the Manager
	DictionaryDatabase *getDictionary();

	//
	// Cache Database Management (if it gets complex, it could
	// be moved to a separate class).  Objects returned are
	// owned by the caller.
	//
	RecordDatabase *createRecordDatabase();
	DocDatabase *createDocDatabase(bt_compare_fcn_type compare = 0);
	docId_t allocateTempDocID();
	
	//
	// Is the current host big-endian?  DB XML's "native"
	// format is little-endian, as of release 2.0
	//
	static bool isBigendian() { return Globals::isBigendian_; }

private:
	Manager(const Manager &o);
	Manager &operator=(const Manager &o);
	void initialize(DbEnv *env);
	void terminate();
	void initTempDbEnv(DbEnv *env);
	
	class ContainerStore {
	public:
		ContainerStore();
		~ContainerStore();
		void initialize(Manager &mgr);

		bool closeContainer(ContainerBase *container,
				    u_int32_t flags);

		// Container ID '0' is only a ContainerBase, but
		// will never be looked up via either of these methods
		XmlContainer findContainer(
			Manager &db,
			const std::string &name,
			Transaction *txn,
			u_int32_t flags,
			XmlContainer::ContainerType type,
			u_int32_t pagesize, u_int32_t seqIncr, int mode,
			bool doVersionCheck);

		XmlContainer findOpenContainer(
			const std::string &name);

		// This method may leave open some races/holes where
		// a container could get deleted out from under an
		// object (see RawNodeValue.cpp).  Keep an eye on this.
		ContainerBase *getContainerFromID(int id,
						  bool acquire) const;

		void releaseRegisteredContainers();
		bool addAlias(const std::string &alias,
			      ContainerBase *container);
		bool removeAlias(const std::string &alias,
				 ContainerBase *container);
	private:
		int insertNewContainer(ContainerBase *container);

		
		typedef std::vector<ContainerBase *> ContainerVector;
		typedef std::map<std::string, int> ContainerIDMap;
		dbxml_mutex_t mutex_;
		ContainerIDMap idmap_;
		ContainerVector containers_;
	};

	DbEnv *dbEnv_;
	bool dbEnvAdopted_;
	DbEnv tempDbEnv_;
	u_int32_t dbEnvOpenFlags_;
	u_int32_t flags_;
	bool autoOpen_;
	std::string dbHome_;

	u_int32_t defaultContainerFlags_;
	u_int32_t defaultPageSize_;
	u_int32_t defaultSequenceIncr_;
	XmlContainer::ContainerType defaultContainerType_;
	ResolverStore *resolvers_;
	int timezone_;

	ContainerStore openContainers_;
	DictionaryDatabase *dictionary_;
	docId_t tempDocId_;

	QueryPlanRoot *defaultQPR_;
	
	friend class XmlContainer;
	friend class Container;
};

}

#endif
