//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002,2008 Oracle.  All rights reserved.
//
// $Id: Modify.cpp,v 1.56 2008/01/08 21:54:23 bostic Exp $
//

#include "DbXmlInternal.hpp"
#include <iostream>
#include "dbxml/XmlException.hpp"
#include "dbxml/XmlQueryExpression.hpp"
#include "dbxml/XmlUpdateContext.hpp"
#include "dbxml/XmlTransaction.hpp"
#include "QueryExpression.hpp"
#include "QueryContext.hpp"
#include "Modify.hpp"
#include "Results.hpp"
#include "Transaction.hpp"
#include "Document.hpp"
#include "Manager.hpp"
#include "nodeStore/NsDocument.hpp"
#include "nodeStore/NsWriter.hpp"
#include "nodeStore/NsEventReader.hpp"

using namespace DbXml;
using namespace std;

// use DB_RMW if transacted and not CDB
#define SET_FLAGS (transaction.isNull() || ((Manager&)((QueryContext&)context).getManager()).isCDBEnv() ? DBXML_LAZY_DOCS : DBXML_LAZY_DOCS|DB_RMW)

//////////////////////////////////
// ModifyStep
//////////////////////////////////

ModifyStep::ModifyStep(const XmlQueryExpression &expr)
	: expr_(expr)
{
	expr_->acquire();
}

ModifyStep::~ModifyStep()
{
	expr_->release();
}

XmlValue ModifyStep::getTypeAsString(XmlModify::XmlObject type)
{
	switch(type) {
	case XmlModify::Element:
		return "element";
	case XmlModify::Attribute:
		return "attribute";
	case XmlModify::Text:
		return "text";
	case XmlModify::ProcessingInstruction:
		return "pi";
	case XmlModify::Comment:
		return "comment";
	}
	return "unknown";
}


//////////////////////////////////
// RemoveStep
//////////////////////////////////

unsigned int RemoveStep::execute(XmlTransaction &transaction, Value *toModify,
	XmlQueryContext &context) const
{
	XmlResults results = expr_->execute(transaction, toModify, context, SET_FLAGS);
	context.setVariableValue("input", results);

	((QueryContext&)context).getManager().query(transaction,
		"declare revalidation skip; "
		"declare variable $input as node()* external; "
		"delete nodes $input",
		context, SET_FLAGS);

	return 1;
}

//////////////////////////////////
// RenameStep
//////////////////////////////////

unsigned int RenameStep::execute(XmlTransaction &transaction, Value *toModify,
	XmlQueryContext &context) const
{
	XmlResults results = expr_->execute(transaction, toModify, context, SET_FLAGS);
	context.setVariableValue("input", results);
	context.setVariableValue("name", name_);

	((QueryContext&)context).getManager().query(transaction,
		"declare revalidation skip; "
		"declare variable $input as node()* external; "
		"declare variable $name as xs:QName external; "
		"for $i in $input return rename node $i as $name",
		context, SET_FLAGS);

	return 1;
}

//////////////////////////////////
// UpdateStep
//////////////////////////////////

unsigned int UpdateStep::execute(XmlTransaction &transaction, Value *toModify,
	XmlQueryContext &context) const
{
	XmlResults results = expr_->execute(transaction, toModify, context, SET_FLAGS);
	context.setVariableValue("input", results);
	context.setVariableValue("content", content_);

	((QueryContext&)context).getManager().query(transaction,
		"declare revalidation skip; "
		"declare variable $input as node()* external; "
		"declare variable $content as xs:string external; "
		"for $i in $input return replace value of node $input with $content",
		context, SET_FLAGS);

	return 1;
}

//////////////////////////////////
// ContentStep
//////////////////////////////////

ContentStep::ContentStep(const XmlQueryExpression &expr, XmlModify::XmlObject type,
	const string &name, XmlResults &content, bool parse)
	: ModifyStep(expr),
	  type_(getTypeAsString(type)),
	  name_(),
	  content_((Results*)content, ((Results&)content).getManager()),
	  parse_(parse)
{
	if(name != "") {
		if(type == XmlModify::Element || type == XmlModify::Attribute)
			name_ = XmlValue(XmlValue::QNAME, name);
		else name_ = XmlValue(XmlValue::STRING, name);
	}
}

static const string contentStepQueryFragment =
"let $toInsert :=\n"
"  if($type = 'element') then (\n"
"    let $parsedContent :=\n"
"      if($parse) then xqilla:parse-xml(concat('<wrapper>', $content, '</wrapper>'))/wrapper/node() else $content\n"
"    return\n"
"      if(exists($name)) then element { $name } { $parsedContent } else $parsedContent\n"
"  ) else if($type = 'attribute') then (\n"
"    attribute { $name } { $content }\n"
"  ) else if($type = 'text') then (\n"
"    text { $content }\n"
"  ) else if($type = 'pi') then (\n"
"    processing-instruction { $name } { $content }\n"
"  ) else if($type = 'comment') then (\n"
"    comment { $content }\n"
"  ) else ()\n"
"return\n";

//////////////////////////////////
// InsertBeforeStep
//////////////////////////////////

unsigned int InsertBeforeStep::execute(XmlTransaction &transaction, Value *toModify,
	XmlQueryContext &context) const
{
	XmlResults results = expr_->execute(transaction, toModify, context, SET_FLAGS);
	context.setVariableValue("input", results);
	context.setVariableValue("type", type_);
	context.setVariableValue("name", name_);

	// Copy the results object for thread safety when iterating
	XmlResults contentResults(new ValueResults(content_,
						   ((Results&)results).getManager()));
	context.setVariableValue("content", contentResults);
	context.setVariableValue("parse", parse_);

	((QueryContext&)context).getManager().query(transaction,
		string("declare revalidation skip;\n"
			"declare variable $input as node()* external;\n"
			"declare variable $type as xs:string external;\n"
			"declare variable $name as item()? external;\n"
			"declare variable $content as item()* external;\n"
			"declare variable $parse as xs:boolean external;\n") +
		contentStepQueryFragment +
		string("for $i in $input return insert nodes $toInsert before $i"),
		context, SET_FLAGS);

	return 1;
}

//////////////////////////////////
// InsertAfterStep
//////////////////////////////////

unsigned int InsertAfterStep::execute(XmlTransaction &transaction, Value *toModify,
	XmlQueryContext &context) const
{
	XmlResults results = expr_->execute(transaction, toModify, context, SET_FLAGS);
	context.setVariableValue("input", results);
	context.setVariableValue("type", type_);
	context.setVariableValue("name", name_);

	// Copy the results object for thread safety when iterating
	XmlResults contentResults(new ValueResults(content_,
						   ((Results&)results).getManager()));
	context.setVariableValue("content", contentResults);
	context.setVariableValue("parse", parse_);

	((QueryContext&)context).getManager().query(transaction,
		string("declare revalidation skip;\n"
			"declare variable $input as node()* external;\n"
			"declare variable $type as xs:string external;\n"
			"declare variable $name as item()? external;\n"
			"declare variable $content as item()* external;\n"
			"declare variable $parse as xs:boolean external;\n") +
		contentStepQueryFragment +
		string("for $i in $input return insert nodes $toInsert after $i"),
		context, SET_FLAGS);

	return 1;
}

//////////////////////////////////
// AppendStep
//////////////////////////////////

unsigned int AppendStep::execute(XmlTransaction &transaction, Value *toModify,
	XmlQueryContext &context) const
{
	XmlResults results = expr_->execute(transaction, toModify, context, SET_FLAGS);
	context.setVariableValue("input", results);
	context.setVariableValue("type", type_);
	context.setVariableValue("name", name_);

	// Copy the results object for thread safety when iterating
	XmlResults contentResults(new ValueResults(content_,
						   ((Results&)results).getManager()));
	context.setVariableValue("content", contentResults);
	context.setVariableValue("parse", parse_);
	context.setVariableValue("location", location_);

	((QueryContext&)context).getManager().query(transaction,
		string("declare revalidation skip;\n"
			"declare variable $input as node()* external;\n"
			"declare variable $type as xs:string external;\n"
			"declare variable $name as item()? external;\n"
			"declare variable $content as item()* external;\n"
			"declare variable $parse as xs:boolean external;\n"
			"declare variable $location as xs:double external;\n") +
		contentStepQueryFragment +
		string("for $i in $input\n"
			"let $before := if($location < 0) then () else $i/node()[$location + 1]\n"
			"return\n"
			"  if(exists($before)) then insert nodes $toInsert before $before\n"
			"  else insert nodes $toInsert as last into $i"),
		context, SET_FLAGS);

	return 1;
}

//////////////////////////////////
// Modify
//////////////////////////////////

Modify::~Modify()
{
	ModifyStep::Vector::iterator end = steps_.end();
	for(ModifyStep::Vector::iterator i = steps_.begin(); i != end; ++i) {
		delete *i;
	}
}

void Modify::addStep(ModifyStep *stepToAdopt)
{
	steps_.push_back(stepToAdopt);
}


unsigned int Modify::execute(XmlTransaction &txn, Value *toModify,
			     XmlQueryContext &context,
			     XmlUpdateContext &uc) const
{
	unsigned int count = 0;

	XmlQueryContext qc(((QueryContext&)context).copy());

	ModifyStep::Vector::const_iterator i = steps_.begin();
	for(; i != steps_.end(); ++i) {
		count += (*i)->execute(txn, toModify, qc);
	}

	return count;
}

unsigned int Modify::execute(XmlTransaction &txn, XmlResults &toModify,
			     XmlQueryContext &context, XmlUpdateContext &uc) const
{
	unsigned int count = 0;

	XmlQueryContext qc(((QueryContext&)context).copy());

	toModify.reset();
	XmlValue value;
	while(toModify.next(value)) {
		ModifyStep::Vector::const_iterator i = steps_.begin();
		for(; i != steps_.end(); ++i) {
			count += (*i)->execute(txn, value, qc);
		}
	}

	return count;
}
