//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002,2008 Oracle.  All rights reserved.
//
// $Id: Modify.hpp,v 1.26 2008/01/08 21:54:23 bostic Exp $
//

#ifndef __MODIFY_HPP
#define __MODIFY_HPP

#include <string>
#include <vector>
#include <dbxml/XmlManager.hpp>
#include <dbxml/XmlModify.hpp>
#include <dbxml/XmlQueryContext.hpp>
#include <dbxml/XmlResults.hpp>
#include "Results.hpp"
#include "ReferenceCounted.hpp"
#include "nodeStore/NsDom.hpp"
#include "UTF8.hpp"

#include <xercesc/parsers/XercesDOMParser.hpp>

namespace DbXml
{

/// ModifyStep base class
class ModifyStep
{
public:
	typedef std::vector<ModifyStep*> Vector;
	
	ModifyStep(const XmlQueryExpression &expr);
	virtual ~ModifyStep();
	
	virtual unsigned int execute(XmlTransaction &transaction, Value *toModify,
		XmlQueryContext &context) const = 0;
	QueryExpression &getExpression() const {
		return *expr_;
	}
	void setExpression(QueryExpression *expr) {
		expr_->release();
		expr_ = expr;
		expr->acquire();
	}

	static XmlValue getTypeAsString(XmlModify::XmlObject type);

private:
	ModifyStep(const ModifyStep &);
	ModifyStep &operator=(const ModifyStep &);
	
protected:
	QueryExpression *expr_;
};

/// Step that removes the given node
class RemoveStep : public ModifyStep
{
public:
	RemoveStep(const XmlQueryExpression &expr) : ModifyStep(expr) {}
	
	virtual unsigned int execute(XmlTransaction &transaction, Value *toModify,
		XmlQueryContext &context) const;
};

/// Step that renames the given node
class RenameStep : public ModifyStep
{
public:
	RenameStep(const XmlQueryExpression &expr, const std::string &name)
		: ModifyStep(expr), name_(XmlValue::QNAME, name) {}
	
	virtual unsigned int execute(XmlTransaction &transaction, Value *toModify,
		XmlQueryContext &context) const;

private:
	XmlValue name_;
};

/// Step that updates the text content of a given node
class UpdateStep : public ModifyStep
{
public:
	UpdateStep(const XmlQueryExpression &expr, const std::string &content)
		: ModifyStep(expr), content_(content) {}
	
	virtual unsigned int execute(XmlTransaction &transaction, Value *toModify,
		XmlQueryContext &context) const;

private:
	XmlValue content_;
};

class ContentStep : public ModifyStep
{
public:
	ContentStep(const XmlQueryExpression &expr, XmlModify::XmlObject type,
		const std::string &name, XmlResults &content, bool parse);

protected:
	XmlValue type_;
	XmlValue name_;
	ValueResults content_;
	XmlValue parse_;
};

/// Step that inserts content before the given node
class InsertBeforeStep : public ContentStep
{
public:
	InsertBeforeStep(const XmlQueryExpression &expr, XmlModify::XmlObject type,
		const std::string &name, XmlResults &content, bool parse)
		: ContentStep(expr, type, name, content, parse) {}

	virtual unsigned int execute(XmlTransaction &transaction, Value *toModify,
		XmlQueryContext &context) const;
};

/// Step that inserts content after the given node
class InsertAfterStep : public ContentStep
{
public:
	InsertAfterStep(const XmlQueryExpression &expr, XmlModify::XmlObject type,
		const std::string &name, XmlResults &content, bool parse)
		: ContentStep(expr, type, name, content, parse) {}

	virtual unsigned int execute(XmlTransaction &transaction, Value *toModify,
		XmlQueryContext &context) const;
};


class AppendStep : public ContentStep
{
public:
	AppendStep(const XmlQueryExpression &expr, XmlModify::XmlObject type,
		const std::string &name, XmlResults &content, bool parse,
		int location)
		: ContentStep(expr, type, name, content, parse),
		  location_((double)location)
	{}

	virtual unsigned int execute(XmlTransaction &transaction, Value *toModify,
		XmlQueryContext &context) const;

private:
	XmlValue location_;
};

class Modify : public ReferenceCountedProtected
{
public:
	Modify(XmlManager &mgr) : mgr_(mgr){}
	virtual ~Modify();

	void addStep(ModifyStep *stepToAdopt);
	
	unsigned int execute(XmlTransaction &txn, Value *toModify,
			     XmlQueryContext &context,
			     XmlUpdateContext &uc) const;
	unsigned int execute(XmlTransaction &txn, XmlResults &toModify,
			     XmlQueryContext &context,
			     XmlUpdateContext &uc) const;

	XmlManager &getManager() { return mgr_; }
	
private:
	mutable XmlManager mgr_;
	
	ModifyStep::Vector steps_;
};

}
#endif
