//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002,2008 Oracle.  All rights reserved.
//
// $Id: QueryExpression.cpp,v 1.96 2008/02/20 17:20:17 gmf Exp $
//

#include "DbXmlInternal.hpp"
#include <sstream>

#include <xqilla/simple-api/XQilla.hpp>
#include <xqilla/simple-api/XQQuery.hpp>
#include <xqilla/context/impl/XQContextImpl.hpp>
#include <xqilla/exceptions/XQException.hpp>
#include <xqilla/exceptions/QueryInterruptedException.hpp>
#include <xqilla/exceptions/QueryTimeoutException.hpp>

#include "PrintDataItemTree.hpp"
#include "QueryExpression.hpp"
#include "QueryContext.hpp"
#include "Results.hpp"
#include "optimizer/ASTReplaceOptimizer.hpp"
#include "optimizer/QueryPlanOptimizer.hpp"
#include "optimizer/ImpliedSchemaGenerator.hpp"
#include "optimizer/QueryPlanGenerator.hpp"
#include "Manager.hpp"
#include "HighResTimer.hpp"

#include "UTF8.hpp"

using namespace DbXml;
using namespace std;

// Counters
#define INCR(ctr) Globals::incrementCounter(ctr)

// QueryExpression
QueryExpression::QueryExpression(const std::string &query, XmlQueryContext &context, Transaction *txn)
	: query_(query),
	  context_(context),
	  qec_(context_, /*debugging*/false),
	  conf_(context, txn, &ci_),
	  xqContext_(XQilla::createContext(XQilla::XQUERY_UPDATE, &conf_, Globals::defaultMemoryManager)),
	  expr_(0)
{
	((Manager &)((QueryContext &)getContext()).getManager())
		.log(Log::C_OPTIMIZER, Log::L_INFO, "Started parse");

	// Set default element namespace if set in context
	const std::string &den = ((QueryContext&)context).getDefaultElementNamespace();
	UTF8ToXMLCh den1(den.c_str(), den.length());
	if (den1.len())
		xqContext_.get()->setDefaultElementAndTypeNS(den1.str());

	HighResTimer t;
	t.start();

	conf_.setMinder(&minder_);
	conf_.setQueryExecutionContext(&qec_);
	conf_.setProjectionInfo(&pi_);

	try {
		((QueryContext&)context_).startQuery();
		expr_ = XQilla::parse(UTF8ToXMLCh(getQuery()).str(), xqContext_, 0,
			XQilla::NO_STATIC_RESOLUTION | XQilla::NO_ADOPT_CONTEXT);

		ScopedPtr<Optimizer> optimizer(createOptimizer(txn));
		optimizer->startOptimize(expr_);
	}
	catch(const QueryInterruptedException &) {
		delete expr_;
		throw XmlException(XmlException::OPERATION_INTERRUPTED,
			"Query was interrupted by the application");
	}
	catch(const QueryTimeoutException &) {
		delete expr_;
		throw XmlException(XmlException::OPERATION_TIMEOUT,
			"Query timed out");
	}
	catch(const XQException &e) {
		delete expr_;
		throw XmlException(XmlException::QUERY_PARSER_ERROR, e);
	}
	catch(...) {
		delete expr_;
		throw;
	}
	t.stop();

	if(Log::isLogEnabled(Log::C_OPTIMIZER, Log::L_INFO)) {
		ostringstream s;
		s << "Finished parse, time taken = " << (t.durationInSeconds() * 1000) << "ms";
		((Manager &)((QueryContext &)getContext()).getManager())
			.log(Log::C_OPTIMIZER, Log::L_INFO, s);
	}
}

QueryExpression::~QueryExpression()
{
	delete expr_;
}

std::string QueryExpression::getQueryPlan() const
{
	return DbXmlPrintAST::print(getCompiledExpression(), xqContext_);
}

DynamicContext *QueryExpression::getDynamicContext() const
{
	return xqContext_;
}

XQQuery *QueryExpression::getCompiledExpression() const
{
	return expr_;
}

bool QueryExpression::isUpdating() const
{
	return expr_->getQueryBody()->getType() == ASTNode::UAPPLY_UPDATES;
}

Results *QueryExpression::execute(Transaction *txn, Value *contextItem, XmlQueryContext &context, u_int32_t flags)
{
	if (contextItem && (contextItem->getType() == XmlValue::BINARY))
		throw XmlException(XmlException::INVALID_VALUE,
				   "XmlQueryExpression::execute: context item cannot be a binary value");

	// Count query
	INCR(Counters::num_queryExec);

	XmlManager &mgr = ((QueryContext&)context).getManager();

	// NOTE: LazyDIResults::next() catches all query exceptions.
	// If Eager resulst are ever changed to not be implemented in terms
	// of Lazy results, that code needs to catch exceptions as well
	// The timer, however, is never used for purely lazy queries, as
	// there is no control over iteration.

	if(isUpdating()) {
		// Auto-transact and execute an updating query
		TransactionGuard txnGuard;
		if(txn != 0 || (flags & DBXML_NO_AUTO_COMMIT) == 0) {
			txn = Transaction::autoTransact(txn, mgr, txnGuard,
				((Manager&)mgr).isTransactedEnv(), ((Manager&)mgr).isCDBEnv());
		}
		flags &= ~DBXML_NO_AUTO_COMMIT;

		{
			XmlResults res(new LazyDIResults(context, contextItem, *this, txn, flags));
			res.hasNext(); // Executes the update
		}

		txnGuard.commit();

		return new ValueResults(mgr);
	}

	Results *ret = new LazyDIResults(context, contextItem, *this, txn, flags & ~DBXML_NO_AUTO_COMMIT);
	if(context.getEvaluationType() == XmlQueryContext::Eager) {
		ret = new ValueResults(ret, mgr, txn);
	}
	return ret;
}

std::map<int, XmlContainer> QueryExpression::getContainers() const
{
	return minder_.getContainers();
}

Optimizer *QueryExpression::createOptimizer(Transaction *txn)
{
	// This needs to be kept in sync with DecisionPointQP::justInTimeOptimize()

// 	OutputTree *before = new OutputTree("Initial tree", xqContext_);
// 	StaticResolver *sr = new StaticResolver(xqContext_, before);
// 	OutputTree *afterStatic = new OutputTree("After Static Resolution", xqContext_, sr);
// 	ASTReplaceOptimizer *rep = new ASTReplaceOptimizer(minder_, xqContext_, afterStatic);
// 	OutputTree *afterRep = new OutputTree("After AST Replace", xqContext_, rep);
// 	StaticTyper *st = new StaticTyper(xqContext_, afterRep);
// 	OutputTree *afterST = new OutputTree("After Static Typing", xqContext_, st);
// 	ImpliedSchemaGenerator *isg = new ImpliedSchemaGenerator(xqContext_, afterST);
// 	OutputTree *afterIsg = new OutputTree("After ImpliedSchemaGenerator", xqContext_, isg);
// 	QueryPlanGenerator *qpGen = new QueryPlanGenerator(xqContext_, afterIsg);
// 	OutputTree *afterQP = new OutputTree("After Query Plan Generation", xqContext_, qpGen);
// 	QueryPlanOptimizer *qpo = new QueryPlanOptimizer(OptimizationContext::RESOLVE_INDEXES, xqContext_, afterQP);
// 	OutputTree *afterQPO = new OutputTree("After QueryPlan Optimisation Phase 1 (RESOLVE_INDEXES)", xqContext_, qpo);
// 	StaticTyper *st2 = new StaticTyper(xqContext_, afterQPO);
// 	OutputTree *afterST2 = new OutputTree("After Static Typing (2)", xqContext_, st2);
// 	QueryPlanOptimizer *qpo2 = new QueryPlanOptimizer(OptimizationContext::REARRANGE, xqContext_, afterST2);
// 	OutputTree *afterQPO2 = new OutputTree("After QueryPlan Optimisation Phase 2 (REARRANGE)", xqContext_, qpo2);
// 	QueryPlanOptimizer *ao = new QueryPlanOptimizer(OptimizationContext::ALTERNATIVES, xqContext_, afterQPO2);
// 	OutputTree *afterAO = new OutputTree("After QueryPlan Optimisation Phase 3 (ALTERNATIVES)", xqContext_, ao);
// 	QueryPlanOptimizer *qpo3 = new QueryPlanOptimizer(OptimizationContext::ADD_STEPS, xqContext_, afterAO);
// 	OutputTree *afterQPO3 = new OutputTree("After QueryPlan Optimisation Phase 4 (ADD_STEPS)", xqContext_, qpo3);
// 	QueryPlanOptimizer *qpo4 = new QueryPlanOptimizer(OptimizationContext::MAKE_PREDICATES, xqContext_, afterQPO3);
// 	OutputTree *afterQPO4 = new OutputTree("After QueryPlan Optimisation Phase 5 (MAKE_PREDICATES)", xqContext_, qpo4);
// 	QueryPlanOptimizer *qpo5 = new QueryPlanOptimizer(OptimizationContext::REMOVE_REDUNDENTS, xqContext_, afterQPO4);
// 	OutputTree *afterQPO5 = new OutputTree("After QueryPlan Optimisation Phase 6 (REMOVE_REDUNDENTS)", xqContext_, qpo5);
// 	StaticTyper *st3 = new StaticTyper(xqContext_, afterQPO5);
// 	OutputTree *afterST3 = new OutputTree("After Static Typing (3)", xqContext_, st3);
// 	ProjectionSchemaAdjuster *psa = new ProjectionSchemaAdjuster(xqContext_, afterST3);
// 	return psa;

	StaticResolver *sr = new StaticResolver(xqContext_);
	ASTReplaceOptimizer *rep = new ASTReplaceOptimizer(minder_, xqContext_, sr);
	StaticTyper *st = new StaticTyper(xqContext_, rep);
	ImpliedSchemaGenerator *isg = new ImpliedSchemaGenerator(xqContext_, st);
	QueryPlanGenerator *qpGen = new QueryPlanGenerator(xqContext_, isg);
	QueryPlanOptimizer *qpo = new QueryPlanOptimizer(OptimizationContext::RESOLVE_INDEXES, xqContext_, qpGen);
	StaticTyper *st2 = new StaticTyper(xqContext_, qpo);
	QueryPlanOptimizer *qpo2 = new QueryPlanOptimizer(OptimizationContext::REARRANGE, xqContext_, st2);
	QueryPlanOptimizer *ao = new QueryPlanOptimizer(OptimizationContext::ALTERNATIVES, xqContext_, qpo2);
	QueryPlanOptimizer *qpo3 = new QueryPlanOptimizer(OptimizationContext::ADD_STEPS, xqContext_, ao);
	QueryPlanOptimizer *qpo4 = new QueryPlanOptimizer(OptimizationContext::MAKE_PREDICATES, xqContext_, qpo3);
	QueryPlanOptimizer *qpo5 = new QueryPlanOptimizer(OptimizationContext::REMOVE_REDUNDENTS, xqContext_, qpo4);
	StaticTyper *st3 = new StaticTyper(xqContext_, qpo5);
	ProjectionSchemaAdjuster *psa = new ProjectionSchemaAdjuster(xqContext_, st3);
	return psa;
}
