//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002,2008 Oracle.  All rights reserved.
//
// $Id: RawNodeValue.cpp,v 1.17 2008/04/02 17:20:34 gmf Exp $
//

#include "DbXmlInternal.hpp"
#include "RawNodeValue.hpp"
#include "Results.hpp"
#include "DocID.hpp"
#include "nodeStore/NsNid.hpp"
#include "nodeStore/NsNode.hpp"
#include "nodeStore/NsDom.hpp"
#include "nodeStore/NsFormat.hpp"
#include "nodeStore/NsWriter.hpp"
#include "nodeStore/NsEventReader.hpp"
#include "Value.hpp"
#include "dataItem/DbXmlNodeImpl.hpp"
#include "dataItem/DbXmlFactoryImpl.hpp"
#include "UTF8.hpp"
#include <dbxml/XmlResults.hpp>

using namespace DbXml;

#define GET_RESULTS() ((Results &) ((XmlResults&)results_))
#define GET_MANAGER() GET_RESULTS().getManager()

// Implementation

RawNodeValue::RawNodeValue()
	: cid_(0), index_(0), type_(0), data_(0), dataSize_(0), returnResults_(0), nodePtr_(0)
{
}

RawNodeValue::RawNodeValue(XmlValue &value, XmlResults &results)
	: results_(results), data_(0), dataSize_(0), returnResults_(0), nodePtr_(0)
{
	if (!value.isNode())
		throw XmlException(XmlException::INVALID_VALUE,
				   "RawNodeValue requires a node");
	DbXmlNodeValue *val = const_cast<DbXmlNodeValue*>(((Value*)value)->
							  getDbXmlNodeValueObject());
	DBXML_ASSERT(val);
	type_ = val->getNodeType();
	const DbXmlNodeImpl *node = val->getNodeImpl();
	DBXML_ASSERT(node || type_ == nsNodeDocument);
	const Document *doc = val->getDocument();
	DBXML_ASSERT(doc);
	did_ = doc->getID();
	cid_ = doc->getContainerID();
	if (node) {
		nid_.copy(node->getNodeID());
		index_ = node->getIndex();
	}else
		nid_.set(rootNid);
}

RawNodeValue::RawNodeValue(const RawNodeValue &other)
	: results_(other.results_), data_(0), dataSize_(0), returnResults_(0), nodePtr_(0)
{
	type_ = other.type_;
	did_ = other.did_;
	cid_ = other.cid_;
	nid_.copy(other.nid_);
	index_ = other.index_;
}

RawNodeValue::RawNodeValue(const void *data, XmlResults &results)
	: results_(results), data_(0), dataSize_(0), returnResults_(0), nodePtr_(0)
{
	loadValue(data);
}

RawNodeValue::~RawNodeValue()
{
	if (data_)
		NsUtil::deallocate(data_);
}

// if the container/did can be found in the
// minder owned by Results, use the DB; otherwise,
// it comes from the Container.
DbWrapper *RawNodeValue::getDocDB() const
{
	Results &res = GET_RESULTS();
	XmlManager &mgr = res.getManager();
	CacheDatabaseMinder &minder = res.getDbMinder();
	CacheDatabase *db = minder.findOrAllocate(mgr, cid_,
						  false /* don't allocate */);
	if (db)
		return db->getDb();

	// get container's DB
	DBXML_ASSERT(cid_ > 0);
	// don't acquire reference on the container
	ContainerBase *cont = ((Manager&)mgr).getContainerFromID(cid_, false);
	if(!cont)
		throw XmlException(XmlException::CONTAINER_CLOSED, "Cannot complete operation because container is closed.");
	return cont->getDbWrapper();
}

DictionaryDatabase *RawNodeValue::getDictDB() const
{
	XmlManager &mgr = GET_MANAGER();
	// don't acquire reference on the container
	ContainerBase *cont = ((Manager&)mgr).getContainerFromID(cid_, false);
	if(!cont)
		throw XmlException(XmlException::CONTAINER_CLOSED, "Cannot complete operation because container is closed.");
	return cont->getDictionaryDatabase();
}

void RawNodeValue::getNsDomNode() const
{
	if (!node_) {
		doc_.init(GET_RESULTS().getOperationContext().txn(),
			  getDocDB(), getDictDB(),
			  did_, cid_, 0);
		NsNode *node = doc_.getNode(nid_);
		// create NsDom object
		
		switch (type_) {
		case nsNodeElement:
		case nsNodeDocument:
			node_ = new NsDomElement(node, &doc_);
			break;
		case nsNodeAttr:
			node_ = new NsDomAttr(node, &doc_, index_);
			break;
		case nsNodeText:
		case nsNodeCDATA:
		case nsNodeComment:
		case nsNodePinst:
			node_ = new NsDomText(node, &doc_, index_);
			break;
		default:
			break;
		}
		DBXML_ASSERT(node_);
	}
}

std::string RawNodeValue::getNodeName() const
{
	if(type_ == nsNodeDocument)
		return "#document";
	getNsDomNode();
	const XMLCh *name =node_->getNsNodeName();
	return XMLChToUTF8(name).str();
}

std::string RawNodeValue::getNodeValue() const
{
	if(type_ == nsNodeDocument)
		return "";
	// TBD -- maybe use utf-8 interfaces for attrs and text
	getNsDomNode();
	const XMLCh *value =node_->getNsNodeValue();
	return XMLChToUTF8(value).str();
}

std::string RawNodeValue::getNamespaceURI() const
{
	if(type_ == nsNodeDocument)
		return "";
	getNsDomNode();
	const char *str = (const char *)node_->getNsUri8();
	if(str)
		return str;
	else
		return "";
}

std::string RawNodeValue::getPrefix() const
{
	if(type_ == nsNodeDocument)
		return "";
	getNsDomNode();
	const char *str = (const char *)node_->getNsPrefix8();
	if(str)
		return str;
	else
		return "";
}

std::string RawNodeValue::getLocalName() const
{
	if(type_ == nsNodeDocument)
		return "#document";
	getNsDomNode();
	const char *str = (const char *)node_->getNsLocalName8();
	if(str)
		return str;
	else
		return "";
}

std::string RawNodeValue::getTypeURI() const
{
	switch(type_) {
	case nsNodeText:
	case nsNodeCDATA:
	case nsNodeAttr:
	case nsNodeElement:
	{
		return "http://www.w3.org/2001/XMLSchema";
	}
	default:
	{
		return "";
	}
	}
}

std::string RawNodeValue::getTypeName() const
{
	if (type_ == nsNodeElement)
		return XMLChToUTF8(DocumentCache::g_szUntyped).str();
	else if (type_ == nsNodeAttr || type_ == nsNodeText || type_ == nsNodeCDATA) {
		return "untypedAtomic";
		//return XMLChToUTF8(ATUntypedAtomic::fgDT_UNTYPEDATOMIC).str();
	}else
		return "";
}

std::string RawNodeValue::asString() const
{
	if(type_ == nsNodeDocument) {
		(void)asDocument();
		DBXML_ASSERT(!xdoc_.isNull());
		std::string content;
		return xdoc_.getContent(content);
	}
	if (type_ == nsNodeElement) {
		// Make output writer
		std::string outstr;
		StringNsStream output(outstr);
		NsWriter writer(&output);

		// Make an event reader
		ScopedPtr<EventReader> reader(
			new NsEventReader(
				GET_RESULTS().getOperationContext().txn(),
				getDocDB(),
				getDictDB(),
				did_, cid_, 0, NS_EVENT_BULK_BUFSIZE, &nid_));
		writer.writeFromReader(*reader);
		return outstr;
	} else if (type_ == nsNodeAttr) {
		getNsDomNode();
		NsDomAttr *attr = (NsDomAttr*)node_.get();
		DBXML_ASSERT(attr);
		std::string str = "{";
		const char *uri = (const char *)attr->getNsUri8();
		if (uri)
			str.append(uri);
		str.append("}");
		str.append((const char *)attr->getNsLocalName8());
		str.append("=\"");
		str.append((const char *)attr->getNsValue8());
		str.append("\"");
		return str;
	} else {
		getNsDomNode();
		NsDomText *text = (NsDomText*)node_.get();
		DBXML_ASSERT(text);
		if(type_ == nsNodeText) {
			return (const char *)text->getNsValue8();
		} else if(type_ == nsNodeComment) {
			std::string str = "<!--";
			str.append((const char *)text->getNsValue8());
			str.append("-->");
			return str;
		} else if(type_ == nsNodeCDATA) {
			std::string str = "<![CDATA[";
			str.append((const char *)text->getNsValue8());
			str.append("]]>");
			return str;
		} else if(type_ == nsNodePinst) {
			std::string str = "<?";
			// PI target is localName
			str.append((const char *)text->getNsNodeName8());
			str.append(" ");
			str.append((const char *)text->getNsValue8());
			str.append("?>");
			return str;
		} else {
			throw XmlException(XmlException::INVALID_VALUE,
					   "cannot create string for type");
		}
	}
	return "";
}

bool RawNodeValue::asBoolean() const
{
	std::string s(asString());
	return (!s.empty() && s.compare("false") != 0);
}

double RawNodeValue::asNumber() const
{
	return strtod(asString().c_str(), 0);
}

std::string RawNodeValue::getNodeHandle() const
{
	if (cid_ == 0) {
		throw XmlException(XmlException::INVALID_VALUE,
				   "getNodeHandle() requires a live node");
	}
	IndexEntry ie;
	ie.setDocID(did_);
	if(type_ != nsNodeDocument) {
		ie.setNodeID(nid_);
		if(type_ == nsNodeElement) {
			ie.setFormat(IndexEntry::NH_ELEMENT_FORMAT);
		} else if(type_ == nsNodeAttr) {
			ie.setFormat(IndexEntry::NH_ATTRIBUTE_FORMAT);
			ie.setIndex(index_);
		} else if(type_ == nsNodeText ||
			type_ == nsNodeCDATA) {
			ie.setFormat(IndexEntry::NH_TEXT_FORMAT);
			ie.setIndex(index_);
		} else if(type_ == nsNodeComment) {
			ie.setFormat(IndexEntry::NH_COMMENT_FORMAT);
			ie.setIndex(index_);
		} else if(type_ == nsNodePinst) {
			ie.setFormat(IndexEntry::NH_PI_FORMAT);
			ie.setIndex(index_);
		} else {
			throw XmlException(XmlException::INVALID_VALUE,
				"Node handle unavailable for node type");
		}
	} else
		ie.setFormat(IndexEntry::NH_DOCUMENT_FORMAT);
	return ie.getNodeHandle();
}

bool RawNodeValue::equals(const Value &other) const
{
	const RawNodeValue *otherVal = other.getRawNodeValueObject();
	if (otherVal) {
		if ((otherVal->type_ == type_) &&
		    (otherVal->nid_ == nid_) &&
		    (otherVal->did_ == did_) &&
		    (otherVal->cid_ == cid_) &&
		    (otherVal->index_ == index_))
			return true;
	}
	return false;
}

RawNodeValue *RawNodeValue::makeRelative(const NsNid &nid,
					 short type, int index) const
{
	RawNodeValue *val = new RawNodeValue(*this);
	val->type_ = type;
	val->index_ = index;
	try{
		val->nid_.copy(nid);
	}catch(XmlException &e){
		delete val;
		throw e;
	}catch(std::exception &e){
		delete val;
		throw e;
	}catch(...){
		delete val;
		throw XmlException(XmlException::INTERNAL_ERROR, "Unexpected internal error");
	}
	return val;
}

XmlValue RawNodeValue::getParentNode() const
{
	if (type_ != nsNodeDocument) {
		getNsDomNode();
		NsDomNodeRef parent = node_->getNsParentNode();
		if (parent)
			return makeRelative(parent->getNodeId(),
					    nsNodeElement, 0);
	}
	return XmlValue();
}

XmlValue RawNodeValue::getChild(bool isFirst) const
{
	getNsDomNode();
	if (node_) {
		NsDomNodeRef child = (isFirst ? node_->getNsFirstChild() :
				      node_->getNsLastChild());
		if (child)
			return makeRelative(child->getNodeId(),
					    child->getNsNodeType(),
					    child->getIndex());
	}
	return XmlValue();
}

XmlValue RawNodeValue::getSibling(bool isNext) const
{
	if (type_ != nsNodeDocument) {
		getNsDomNode();
		if (node_) {
			NsDomNodeRef sib = (isNext ? node_->getNsNextSibling() :
					    node_->getNsPrevSibling());
			if (sib)
				return makeRelative(sib->getNodeId(),
						    sib->getNsNodeType(),
						    sib->getIndex());
		}
	}
	return XmlValue();
}

XmlValue RawNodeValue::getOwnerElement() const {

	if (type_ ==  nsNodeAttr)
		return makeRelative(nid_, nsNodeElement, 0);
	throw XmlException(XmlException::INVALID_VALUE,
			   "Node is not an attribute node");

}

XmlResults RawNodeValue::getAttributes() const {
	XmlManager &mgr = GET_MANAGER();
	ValueResults *vr = 0;
	try{
		vr = new ValueResults(mgr, getTransaction());
		if (type_ == nsNodeElement) {
			getNsDomNode();
			NsDomElement *elem = (NsDomElement*)node_.get();
			DBXML_ASSERT(elem);
			int size = elem->getNumAttrs();
			for(int i = 0; i < size; ++i){
				RawNodeValue *att = makeRelative(nid_, nsNodeAttr, i);
				att->copyContext(cid_, did_, results_, xdoc_, returnResults_);
				vr->add(att);
			}
		}
	}catch(XmlException &e){
		if(vr) delete vr;
		throw e;
	}catch(std::exception &e){
		if(vr) delete vr;
		throw e;
	}catch(...){
		if(vr) delete vr;
		throw XmlException(XmlException::INTERNAL_ERROR, "Unexpected internal error");
	}
	return vr;
}

XmlResults RawNodeValue::getAttributes(XmlResults *result) const {
	XmlManager &mgr = GET_MANAGER();
	ValueResults *vr = 0;
	try{
		vr = new ValueResults(mgr, getTransaction());
		if (type_ == nsNodeElement) {
			getNsDomNode();
			NsDomElement *elem = (NsDomElement*)node_.get();
			DBXML_ASSERT(elem);
			int size = elem->getNumAttrs();
			for(int i = 0; i < size; ++i){
				RawNodeValue *att = makeRelative(nid_, nsNodeAttr, i);
				if(returnResults_ != 0)
					att->copyContext(cid_, did_, results_, xdoc_, returnResults_);
				else{
					att->copyContext(cid_, did_, results_, xdoc_, (uint64_t)result);
				}
				vr->add(att);
			}
		}
	}catch(XmlException &e){
		if(vr) delete vr;
		throw e;
	}catch(std::exception &e){
		if(vr) delete vr;
		throw e;
	}catch(...){
		if(vr) delete vr;
		throw XmlException(XmlException::INTERNAL_ERROR, "Unexpected internal error");
	}
	return vr;
}

XmlEventReader &RawNodeValue::asEventReader() const
{
	const NsNid *nid = (type_ == nsNodeElement ? &nid_ : 0);
	if ((type_ == nsNodeDocument) ||
	    (type_ == nsNodeElement)) {
		return *(new NsEventReader(
				 GET_RESULTS().getOperationContext().txn(),
				 getDocDB(),
				 getDictDB(),
				 did_, cid_, 0, NS_EVENT_BULK_BUFSIZE, nid));
	}
	// The only way to support asEventReader on non-elements is if
	// the XmlEventReader interface were able to iterate through attributes
	// one at a time, rather than using an indexed interface.
	throw XmlException(XmlException::INVALID_VALUE,
			   "XmlValue::asEventReader requires an element node");
}

const XmlDocument &RawNodeValue::asDocument() const
{
	if (xdoc_.isNull()) {
		XmlManager &mgr = GET_MANAGER();
		xdoc_ = mgr.createDocument();
		Document *doc = (Document*)xdoc_;
		doc->setID(did_);
		doc->setContainerID(cid_);
		doc->setTransaction(GET_RESULTS().getOperationContext().txn());
		if(cid_ == 0)
			doc->setDbMinder(GET_RESULTS().getDbMinder());
		else
			doc->setAsNotMaterialized();
	}
	return xdoc_;
}

void RawNodeValue::copyContext(int cid, const DocID did, XmlResults results, 
							   XmlDocument xdoc, uint64_t returnResults)
{
	cid_ = cid;
	did_ = did;
	results_ = results;
	xdoc_ = xdoc;
	returnResults_ = returnResults;
}

DbXmlNodeImpl *RawNodeValue::getNodeImpl(DynamicContext *context)
{
	if(!context || nid_.isNull() || type_ == nsNodeDocument)
		return 0;
	if(nodePtr_.isNull()){
		getNsDomNode();
		nodePtr_ = ((DbXmlFactoryImpl*)context->getItemFactory())->createNode(node_.get(), asDocument(), context);
	}
	return (DbXmlNodeImpl*)nodePtr_.get();
}

// Order is:
// doc ID
// container id
// type
// index
// nid
int RawNodeValue::marshal(void *&data)
{
	dataSize_ = did_.marshalSize();
	dataSize_ += NsFormat::countInt(cid_);
	dataSize_ += 1; // type
	dataSize_ += NsFormat::countInt(index_);
	if (type_ != nsNodeDocument)
		dataSize_ += NsFormat::countId(nid_);
	dataSize_ += 4; // add padding
	data_ = NsUtil::allocate(dataSize_, "RawNodeValue dump");
	xmlbyte_t *ptr = (xmlbyte_t *)data_;
	ptr += did_.marshal((void*)ptr);
	ptr += NsFormat::marshalInt(ptr, cid_);
	ptr += NsFormat::marshalInt(ptr, type_);
	ptr += NsFormat::marshalInt(ptr, index_);
	if (type_ != nsNodeDocument)
		NsFormat::marshalId(ptr, nid_);
	data = data_;
	return dataSize_;
}

int RawNodeValue::dumpValue(void *&data)
{
	if (!data_)
		return marshal(data);
	data = data_;
	return dataSize_;
}

void RawNodeValue::loadValue(const void *data)
{
	const xmlbyte_t *ptr = (const xmlbyte_t *)data;
	ptr += did_.unmarshal((const void*)ptr);
	ptr += NsFormat::unmarshalInt(ptr, (int32_t*)&cid_);
	int32_t ttype;
	ptr += NsFormat::unmarshalInt(ptr, &ttype);
	type_ = (short) ttype;
	ptr += NsFormat::unmarshalInt(ptr, (int32_t*)&index_);
	if (type_ != nsNodeDocument)
		nid_.set((const unsigned char *)ptr);
	else
		nid_.set(rootNid);
}

Transaction *RawNodeValue::getTransaction() const
{
	Results &res = GET_RESULTS();
	return res.getOperationContext().txn();
}
