//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002,2008 Oracle.  All rights reserved.
//
// $Id: RawNodeValue.hpp,v 1.8 2008/02/14 20:26:47 lauren Exp $
//

#ifndef __RAWNODEVALUE_H
#define	__RAWNODEVALUE_H

/*
 * An implementation of NodeValue (Value) that does not maintain
 * state in the form of references to nodes and documents.
 * It holds container, document and node IDs and resolves these
 * in the context of an XmlResults object when necessary.
 *
 * Actual nodes may be directly in container databases or in temporary
 * databases associated with Results.
 *
 * For the time being (2.4), this class is ONLY used by the Java API.
 * If it weren't for Windows export issues, it would reside entirely
 * in Java-only code (and probably *not* implement NodeValue).
 *
 * It could, however, replace DbXmlNodeValue if desired.  In order
 * to do so, it'd need to implement standalone attribute and text nodes,
 * which are currently handled in DbXmlNodeImpl classes.
 */
#include "Value.hpp"
#include <dbxml/XmlResults.hpp>
#include <dbxml/XmlPortability.hpp>
#include "nodeStore/NsDoc.hpp"

namespace DbXml
{

class DbWrapper;
class DictionaryDatabase;
	
class DBXML_EXPORT RawNodeValue : public NodeValue {
public:
	RawNodeValue();
	RawNodeValue(XmlValue &value, XmlResults &results);  // ctor from NodeValue
	RawNodeValue(const void *data, XmlResults &results); // ctor from dumped content
	RawNodeValue(const RawNodeValue &other);
	~RawNodeValue();

	std::string getTypeURI() const;
	std::string getTypeName() const;

	virtual double asNumber() const;
	virtual std::string asString() const;
	virtual bool asBoolean() const;
	virtual const XmlDocument &asDocument() const; // note: will throw for now
	virtual XmlEventReader &asEventReader() const;
	virtual bool equals(const Value &v) const;

	/// Return an opaque handle to the node
	std::string getNodeHandle() const;

	std::string getNodeName() const;
	std::string getNodeValue() const;
	std::string getNamespaceURI() const;
	std::string getPrefix() const;
	std::string getLocalName() const;
	short getNodeType() const { return type_; }
	XmlValue getParentNode() const;
	XmlValue getFirstChild() const { return getChild(true); }
	XmlValue getLastChild() const { return getChild(false); }
	XmlValue getPreviousSibling() const { return getSibling(false); }
	XmlValue getNextSibling() const { return getSibling(true); }
	XmlResults getAttributes() const;
	XmlResults getAttributes(XmlResults *result) const;
 	XmlValue getOwnerElement() const;

	virtual const RawNodeValue * getRawNodeValueObject() const { return this; }	
	virtual DbXmlNodeImpl *getNodeImpl(DynamicContext *context = 0);

	void copyContext(int cid, const DocID did, XmlResults results, XmlDocument xdoc, uint64_t returnResults);

	// Methods used by Java API
	const DocID &getDocumentID() const { return did_; }
	int dumpValue(void *&data); // returns data and size of data
	void setDocument(XmlDocument &doc) { xdoc_ = doc; }
	void setReturnResults(uint64_t results) { returnResults_ = results; }
	uint64_t getReturnResults() { return returnResults_; }
	Transaction *getTransaction() const;
private:
	// internal methods
	DbWrapper *getDocDB() const;
	DictionaryDatabase *getDictDB() const;
	void getNsDomNode() const;
	void loadValue(const void *data);
	int marshal(void *&data);
	RawNodeValue *makeRelative(const NsNid &nid, short type, int index) const;
	XmlValue getChild(bool isFirst) const;
	XmlValue getSibling(bool isNext) const;
private:
	// state
	DocID did_;     // doc ID
	int cid_;       // container ID
	int index_;     // index for attribute or text types
	short type_; // node type
	NsNidWrap nid_; // nid
	mutable XmlResults results_;
	void *data_;
	int dataSize_;
	mutable NsDomNodeRef node_;
	mutable NsDoc doc_;
	mutable XmlDocument xdoc_;
	uint64_t returnResults_; // Set the Java XmlResults to this pointer
	Node::Ptr nodePtr_;
};

}

#endif
