//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002,2008 Oracle.  All rights reserved.
//
// $Id: Map.cpp,v 1.7 2008/01/08 21:54:23 bostic Exp $
//

#include "DbXmlInternal.hpp"
#include "Map.hpp"

#include <xqilla/context/DynamicContext.hpp>
#include <xqilla/context/ContextHelpers.hpp>
#include <xqilla/runtime/SequenceResult.hpp>
#include <xqilla/ast/XQNav.hpp>
#include <xqilla/ast/XQPredicate.hpp>
#include <xqilla/context/VariableTypeStore.hpp>
#include <xqilla/utils/XPath2Utils.hpp>

using namespace DbXml;
using namespace std;

Map::Map(ASTNode *arg1, ASTNode *arg2, const XMLCh *uri, const XMLCh *name, XPath2MemoryManager* memMgr)
	: DbXmlASTNode(MAP, memMgr),
	  arg1_(arg1),
	  arg2_(arg2),
	  uri_(uri),
	  name_(name),
	  varSrc_(memMgr)
{
	const StaticAnalysis &arg1Src = arg1_->getStaticAnalysis();
	_src.add(arg1Src);

	const StaticAnalysis &arg2Src = arg2_->getStaticAnalysis();

	if(name_ == 0) {
		_src.addExceptContextFlags(arg2Src);
	} else {
		StaticAnalysis newSrc(memMgr);
		newSrc.add(arg2Src);
		newSrc.removeVariable(uri_, name_);
		_src.add(newSrc);
	}

	_src.getStaticType() = arg2Src.getStaticType();
	_src.setProperties(XQNav::combineProperties(arg1Src.getProperties(), arg2Src.getProperties()));
}

ASTNode *Map::staticTyping(StaticContext *context)
{
	VariableTypeStore* varStore=context->getVariableTypeStore();

	_src.clear();

	arg1_ = arg1_->staticTyping(context);
	const StaticAnalysis &arg1Src = arg1_->getStaticAnalysis();
	_src.add(arg1Src);

	varSrc_.getStaticType() = arg1Src.getStaticType();
	varSrc_.setProperties(StaticAnalysis::DOCORDER | StaticAnalysis::GROUPED |
		StaticAnalysis::PEER | StaticAnalysis::SUBTREE | StaticAnalysis::SAMEDOC |
		StaticAnalysis::ONENODE | StaticAnalysis::SELF);

	StaticType oldContextItemType = context->getContextItemType();
	if(name_ == 0) {
		context->setContextItemType(varSrc_.getStaticType());
	} else {
		varStore->addLogicalBlockScope();
		varStore->declareVar(uri_, name_, varSrc_);
	}

	arg2_ = arg2_->staticTyping(context);
	const StaticAnalysis &arg2Src = arg2_->getStaticAnalysis();

	StaticAnalysis newSrc(context->getMemoryManager());
	if(name_ == 0) {
		newSrc.addExceptContextFlags(arg2Src);
		context->setContextItemType(oldContextItemType);
	} else {
		newSrc.add(arg2Src);
		newSrc.removeVariable(uri_, name_);
		varStore->removeScope();
	}

	_src.add(newSrc);
	_src.getStaticType() = arg2Src.getStaticType();
	_src.setProperties(XQNav::combineProperties(arg1Src.getProperties(), arg2Src.getProperties()));

	return this;
}

DbXmlResult Map::createResult(const DbXmlResult &contextItems, unsigned int props,
	DynamicContext *context) const
{
	DbXmlResult result = createDbXmlResult(arg1_, context);

	if(name_ != 0) {
		return new MapResult(result, arg2_, uri_, name_);
	}

	DbXmlASTNode *dbxmlast = DbXmlASTNode::getDbXmlASTNode(arg2_);
	if(arg2_->getStaticAnalysis().isContextSizeUsed()) {
		// We need the context size, so convert to a Sequence to work it out
		Sequence seq(result->toSequence(context));
		result =  new NavStepResult(new SequenceResult(this, seq), arg2_, seq.getLength());
	} else if(dbxmlast != 0) {
		result = dbxmlast->createResult(result, arg1_->getStaticAnalysis().getProperties(), context);
	} else {
		result = new NavStepResult(result, arg2_, 0);
	}

	return result;
}

////////////////////////////////////////////////////////////////////////////////////////////////////

MapResult::MapResult(const Result &parent, const ASTNode *step, const XMLCh *uri, const XMLCh *name)
	: ResultImpl(step),
	  parent_(parent),
	  step_(step),
	  uri_(uri),
	  name_(name),
	  stepResult_(0),
	  item_(0),
	  scope_(0)
{
}

Result MapResult::getVar(const XMLCh *namespaceURI, const XMLCh *name) const
{
	if(XPath2Utils::equals(name, name_) && XPath2Utils::equals(namespaceURI, uri_))
		return item_;

	return scope_->getVar(namespaceURI, name);
}

Item::Ptr MapResult::next(DynamicContext *context)
{
	if(scope_ == 0) {
		scope_ = context->getVariableStore();
	}

	AutoVariableStoreReset reset(context, this);

	Item::Ptr result;
	while((result = stepResult_->next(context)).isNull()) {
		context->testInterrupt();

		context->setVariableStore(scope_);
		item_ = parent_->next(context);
		if(item_.isNull()) {
			parent_ = 0;
			return 0;
		}

		context->setVariableStore(this);
		stepResult_ = step_->createResult(context);
	}

	return result;
}

////////////////////////////////////////////////////////////////////////////////////////////////////

DbXmlPredicate::DbXmlPredicate(ASTNode *expr, ASTNode *predicate, const XMLCh *uri, const XMLCh *name, XPath2MemoryManager *memMgr)
	: DbXmlASTNode(DBXML_PREDICATE, memMgr),
	  expr_(expr),
	  pred_(predicate),
	  uri_(uri),
	  name_(name),
	  varSrc_(memMgr)
{
	_src.copy(expr_->getStaticAnalysis());

	if(name_ == 0) {
		_src.addExceptContextFlags(pred_->getStaticAnalysis());
	} else {
		StaticAnalysis newSrc(memMgr);
		newSrc.add(pred_->getStaticAnalysis());
		newSrc.removeVariable(uri_, name_);
		_src.add(newSrc);
	}
}

ASTNode *DbXmlPredicate::staticTyping(StaticContext *context)
{
	VariableTypeStore* varStore=context->getVariableTypeStore();

	_src.clear();

	expr_ = expr_->staticTyping(context);
	const StaticAnalysis &exprSrc = expr_->getStaticAnalysis();
	_src.copy(exprSrc);

	varSrc_.getStaticType() = exprSrc.getStaticType();
	varSrc_.setProperties(StaticAnalysis::DOCORDER | StaticAnalysis::GROUPED |
		StaticAnalysis::PEER | StaticAnalysis::SUBTREE | StaticAnalysis::SAMEDOC |
		StaticAnalysis::ONENODE | StaticAnalysis::SELF);

	StaticType oldContextItemType = context->getContextItemType();
	if(name_ == 0) {
		context->setContextItemType(varSrc_.getStaticType());
	} else {
		varStore->addLogicalBlockScope();
		varStore->declareVar(uri_, name_, varSrc_);
	}

	pred_ = pred_->staticTyping(context);
	const StaticAnalysis &predSrc = pred_->getStaticAnalysis();

	StaticAnalysis newSrc(context->getMemoryManager());
	if(name_ == 0) {
		newSrc.addExceptContextFlags(predSrc);
		context->setContextItemType(oldContextItemType);
	} else {
		newSrc.add(predSrc);
		newSrc.removeVariable(uri_, name_);
		varStore->removeScope();
	}

	_src.add(newSrc);

	return this;
}

DbXmlResult DbXmlPredicate::createResult(const DbXmlResult &contextItems, unsigned int props, DynamicContext *context) const
{
	DbXmlResult result = createDbXmlResult(expr_, context);

	if(name_ != 0) {
		return new VarPredicateResult(result, pred_, uri_, name_);
	}

	return new PredicateFilterResult(result, pred_, 0);
}

////////////////////////////////////////////////////////////////////////////////////////////////////

VarPredicateResult::VarPredicateResult(const Result &parent, const ASTNode *pred, const XMLCh *uri, const XMLCh *name)
	: ResultImpl(pred),
	  parent_(parent),
	  pred_(pred),
	  uri_(uri),
	  name_(name),
	  item_(0),
	  scope_(0)
{
}

Result VarPredicateResult::getVar(const XMLCh *namespaceURI, const XMLCh *name) const
{
	if(XPath2Utils::equals(name, name_) && XPath2Utils::equals(namespaceURI, uri_))
		return item_;

	return scope_->getVar(namespaceURI, name);
}

Item::Ptr VarPredicateResult::next(DynamicContext *context)
{
	if(scope_ == 0) {
		scope_ = context->getVariableStore();
	}

	AutoVariableStoreReset reset(context, scope_);

	while((item_ = parent_->next(context)).notNull()) {
		context->testInterrupt();

		context->setVariableStore(this);
		if(pred_->createResult(context)->getEffectiveBooleanValue(context, this))
			break;

		context->setVariableStore(scope_);
	}

	return item_;
}

////////////////////////////////////////////////////////////////////////////////////////////////////

DbXmlNodeCheck::DbXmlNodeCheck(ASTNode *arg, XPath2MemoryManager* memMgr)
	: DbXmlASTNode(NODE_CHECK, memMgr),
	  arg_(arg)
{
	_src.copy(arg_->getStaticAnalysis());
	_src.getStaticType().typeIntersect(StaticType::NODE_TYPE);
}

ASTNode *DbXmlNodeCheck::staticTyping(StaticContext *context)
{
	_src.clear();

	arg_ = arg_->staticTyping(context);
	_src.copy(arg_->getStaticAnalysis());

	if(_src.getStaticType().isType(StaticType::NODE_TYPE)) {
		return arg_;
	}

	_src.getStaticType().typeIntersect(StaticType::NODE_TYPE);

	return this;
}

DbXmlResult DbXmlNodeCheck::createResult(const DbXmlResult &contextItems, unsigned int props,
	DynamicContext *context) const
{
	return new IntermediateStepCheckResult(this, arg_->createResult(context));
}

////////////////////////////////////////////////////////////////////////////////////////////////////

DbXmlLastStepCheck::DbXmlLastStepCheck(ASTNode *arg, XPath2MemoryManager* memMgr)
	: DbXmlASTNode(LAST_STEP_CHECK, memMgr),
	  arg_(arg)
{
	_src.copy(arg_->getStaticAnalysis());
}

ASTNode *DbXmlLastStepCheck::staticTyping(StaticContext *context)
{
	_src.clear();

	arg_ = arg_->staticTyping(context);
	_src.copy(arg_->getStaticAnalysis());

	if(!_src.getStaticType().containsType(StaticType::NODE_TYPE) ||
		!_src.getStaticType().containsType(StaticType::ANY_ATOMIC_TYPE)) {
		return arg_;
	}

	return this;
}

DbXmlResult DbXmlLastStepCheck::createResult(const DbXmlResult &contextItems, unsigned int props,
	DynamicContext *context) const
{
	return new LastStepCheckResult(this, arg_->createResult(context));
}

