//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002,2008 Oracle.  All rights reserved.
//
// $Id: NsHandlerBase.cpp,v 1.36 2008/01/16 20:55:25 gmf Exp $
//

#include "NsHandlerBase.hpp"
#include "NsEvent.hpp"
#include "NsConstants.hpp"

using namespace DbXml;


NsHandlerBase::NsHandlerBase(NsDocument *doc,
			     EventWriter *ewriter)
	: _doc(doc), _ewriter(ewriter),
	  _current(0), _previous(0), _depth(0),
	  _lastWasText(false), _textList(0)
{
}

NsHandlerBase::~NsHandlerBase()
{
	NsNode *node;
	while(_current) {
		node = _current;
		_current = node->getParent();
		node->release();
	}
	if(_previous) {
		_previous->release();
	}
	if (_textList)
		NsNode::freeTextList(_textList);
}

void
NsHandlerBase::nextId(NsFullNid *id)
{
	_nidGen.nextId(id);
}

/*
 * Generic parts of node creation --
 */
void
NsHandlerBase::startElem(NsNode *node,
			 const xmlbyte_t *baseName,
			 const char *uri, // may be null
			 const char *prefix,
			 bool isDonated)
{
	int uriIndex = 0;
	int prefixIndex = NS_NOPREFIX;
	//
	// Structural initialization
	//
	++_depth;
	node->setParent(_current);
	node->setLevel(_depth);
	if (_previous) {
		node->setNextPrev(_previous);
		NsNode *tprev = _previous;
		_previous = 0;
		_doc->completeNode(tprev);
	}
	
	//
	// element-specific initialization
	//
	//
	// Update namespace info
	//
	if (uri) {
		uriIndex = _doc->addIDForString(uri, ::strlen(uri));
		node->setFlag(NS_HASURI);
		// cannot be prefix without uri
		if (prefix) {
			prefixIndex = _doc->addIDForString(prefix,
							   ::strlen(prefix));
			node->setFlag(NS_NAMEPREFIX);
		}
	}
	
	// name
	node->setUriIndex(uriIndex);
	node->setNamePrefix(prefixIndex);
	size_t nlen = NsUtil::nsStringLen(baseName);
	NsNode::createText(node->getNameText(),
			   baseName, nlen, isDonated);

	//
	// conditional initialization
	//
	if (_current) {
		DBXML_ASSERT(node->getParent() == _current);

		// tell node we're about to start it up.  It's responsible
		// for copying parent/child NIDs if necessary
		_textList = _current->startElem(node, _textList);
	}
	_lastWasText = false;

	// reset current, parse-global state
	_current = node;

	// addition of attributes and calling of event handler
	// are done in the calling instance
}

void
NsHandlerBase::end()
{
	endElem();
	if (_previous) {
		NsNode *tprev = _previous;
		_previous = 0;
		_doc->completeNode(tprev);
	}
}

void
NsHandlerBase::endDoc()
{
	end();
	_doc->completeDocument();
}

void
NsHandlerBase::endElem()
{
	// don't lose track of _current until safe from exceptions...
	NsNode *node = _current;
	
	// tell node that we're about to be done with it, passing
	// parent along
	_textList = node->endElem(node->getParent(), _previous, _textList);
	if (_previous) {
		NsNode *tprev = _previous;
		_previous = 0;
		_doc->completeNode(tprev); // can throw deadlock
	}
	_previous = node;
	_current = node->getParent();

	// pop depth counter, but not if was doc node
	// which means _current is null
	if (_current)
		--_depth;

	// must reset current text after element ends
	_lastWasText = false;
}

//
// Add text to the current text list, allocating one if necessary
//  text -- pointer to text
//  len -- length in bytes (utf-8), not including null.
//  textType -- type of text, plus, optionally NS_ENTITY_CHK flag (means
//    checks must use logical operators).
//  isDonated -- true if text string is donated memory
//
void
NsHandlerBase::addText(const xmlbyte_t *text, size_t len,
		       uint32_t textType, bool isDonated)
{
	DBXML_ASSERT(_current);
	if (_lastWasText && (nsTextType(textType) == NS_TEXT)) {
		/* can coalesce */
		NsNode::coalesceText(_textList,
				     text, len, isDonated);
		if (isDonated) {
			NsUtil::deallocate((void*)text);
		}
	} else {
		// textType may include NS_ENTITY_CHK, if donated.
		_textList = _current->addText(_textList, text, len,
					      textType, isDonated);
		_lastWasText = (nsTextType(textType) == NS_TEXT);
	}
	if (textType & NS_ENTITY_CHK)
		_textList->tl_text[_textList->tl_ntext - 1].te_type |=
			NS_ENTITY_CHK;
}

//
// Add processing instruction to the current node.
//  target -- pointer to target text
//  data -- pointer to data text
//  isDonated -- true if text string is donated memory
//  len -- length in bytes (utf-8) of both strings, including null.
//
void
NsHandlerBase::addPI(const xmlbyte_t *target, const xmlbyte_t *data,
		     bool isDonated, size_t len)
{
	DBXML_ASSERT(_current);

	_textList = _current->addPI(_textList, target, data,
				    isDonated, len);

	_lastWasText = false;
}
