//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002,2008 Oracle.  All rights reserved.
//
// $Id: NsNid.cpp,v 1.26 2008/03/18 21:00:50 gmf Exp $
//

/**
 * NsFullNid.cpp
 *
 * Node ID and accessors and nid generation
 *
 * This includes both transient and persistent formats of
 * this object, encapsulating its format, allowing for
 * possible changes
 */

#include "NsNid.hpp"
#include "NsConstants.hpp"
#include "NsUtil.hpp"
#include "NsDom.hpp" // for NsDomInsertType enum
#include <sstream>

XERCES_CPP_NAMESPACE_USE
using namespace DbXml;

/**
 * Node Id Format
 *
 * Node Ids are stored as null-terminated arrays of bytes,
 * with valid values as determined by the NS_ID_* constants
 * defined below.  Collation order is simple string collation.
 *
 * Each ID is of the form:
 *  byte 0:  number of bytes before the "decimal" point (n)
 *  bytes 1-(n-1): the primary node id
 *  bytes n-end_of_string: bytes after the "decimal" point,
 *    allocated/used when inserting new node ids between other
 *    node ids.  An unmodified document will not have any of
 *    these
 *
 * For example, let's say that the valid characters are only a-z.
 * In this case, a newly-created document will have ids like this:
 * 1 a 0
 * 1 b 0
 * ...
 * 1 z 0
 * 2 aa 0
 * 2 aa 0
 * ...
 *
 * This format is designed for these requirements:
 *  1.  Be compact when there are no node insertions after creation
 *  2.  Allow efficient comparison
 *  3.  Allow efficient insertion of new node ids between existing nodes
 *
 * Here's an example or two of inserting new nodes, using the
 * same a-z range as above:
 * 1.  Between "1a0" and "1b0" ==> 1am
 * 2.  Between "1a0" and "1am" ==> 1ag
 * 3.  Between "1a0" and "1ab" ==> 1aam (cannot use "1aa")
 */

const NsFullNid NsFullNid::docRootFullNid = NsFullNid();
const NsNid NsNid::docRootNid;
const NsNid NsNid::docMetaDataNid;

#define NID_BETW_INITIAL_SIZE 4 // must be two or more
#define NID_INITIAL_DIGIT 0

#ifdef DEBUG
// Add conditional that can be turned on to test behavior
// of large node IDs.  This can be hard to trigger with content
// alone.  NEVER use this in production or even normal debug
// builds...
//#define USE_LARGE_NIDS 1
#endif

//
// Node Id Comparison function
//
// A simple lexicographical comparison, stopping at the first null.
//
extern "C" {
int
nsCompareNodes(DB *db, const DBT *dbt1, const DBT *dbt2)
{
	// Assume null termination of node ids
	int res;
	const xmlbyte_t *p1 = (const xmlbyte_t *)dbt1->data;
	const xmlbyte_t *p2 = (const xmlbyte_t *)dbt2->data;
	while((res = (int)*p1 - (int)*p2) == 0 &&
	      *p1 != 0) {
		++p1; ++p2;
	}
	return res;
}
};

// NsFullNid methods

static inline void _nextId(unsigned char *id, uint32_t &digit)
{
	unsigned char *dest = id + 1;
	int32_t i = digit;
	while (i >= 0 && (dest[i] == NS_ID_LAST)) {
		dest[i--] = NS_ID_FIRST;
	}
	if (i < 0) {
		++digit;
		dest[digit] = NS_ID_FIRST;
	} else
		dest[i] += 1;

	DBXML_ASSERT(dest[digit+1] == 0);
	// assign number of digits to leading byte
	*id = (unsigned char)(digit + 1);
}

static inline xmlbyte_t idbyte2num(xmlbyte_t idbyte)
{
	return (idbyte == 0 || idbyte == NS_ID_ZERO) ? 0 :
		idbyte + 1 - NS_ID_FIRST;
}

static inline xmlbyte_t num2idbyte(xmlbyte_t num)
{
	return num == 0 ? NS_ID_ZERO : num + NS_ID_FIRST - 1;
}

static inline xmlbyte_t nextIdbyteAsNum(const xmlbyte_t *&p)
{
	if(*p == 0) return 0;
	++p;
	return idbyte2num(*p);
}

// TBD GMF -- when doing up date redefine "insertAfter"
// in a public place
#define insertAfter 2
#define POW_OF_2 3
static inline int avgbyte(int pb, int nb, int itype)
{
	int avg = (pb + nb) >> 1;
	for (int i = 0; avg != pb && i < POW_OF_2; i++) {
		if (itype == insertAfter)
			avg = ((int)nb + (int)avg) >> 1;
		else
			avg = ((int)pb + (int)avg) >> 1;
	}
	return avg;
}


// generate id for this object bewteen prev and next
void
NsFullNid::getBetweenNid(const NsNid *prev,
			 const NsNid *next,
			 int itype)
{
	DBXML_ASSERT(prev);

	freeNid(); // free any existing ID

	uint32_t plen = prev->getLen();
	uint32_t nlen = next ? next->getLen() : 0;
	if(plen == 0 && nlen == 0) {
		memset(this, 0, sizeof(NsFullNid));
		return;
	}

	// maximum length of the between id is one
	// greater than the maximum length of the
	// prev or next id
	uint32_t maxlen = plen > nlen ? plen : nlen;
	maxlen += 1;

	// Allocate memory to build the id in
	xmlbyte_t *dest;
	if (maxlen > NID_BYTES_SIZE) {
		dest = (xmlbyte_t*)
			::malloc(maxlen * sizeof(xmlbyte_t) + 1);
		nidStore.nidPtr = dest;
	} else {
		dest = nidStore.nidBytes;
	}

	if (nlen == 0) {
		// no next, so we must be appending to the end of
		// the document, just add a next nid to previous
		memcpy(dest, prev->getBytes(), plen);
		dest[maxlen] = 0;

		// eliminate everything after the decimal, and
		// allow for an extra digit
		dest[dest[0]+1] = 0;
		dest[dest[0]+2] = 0;

		// idDigit is 0-based, subtract one from num bytes
		uint32_t tdig = dest[0] - 1;
		_nextId(dest, tdig);
		nidLen = (uint32_t)::strlen((const char *)dest) + 1;
		if (nidLen > NID_BYTES_SIZE)
			setLenAlloc(nidLen);
		else if (maxlen > NID_BYTES_SIZE) {
			// maintain invariant that smaller NIDs are
			// in nidBytes array.
			memcpy(nidStore.nidBytes, dest, nidLen);
			::free((void*)dest);
		}
		return;
	}
	
	xmlbyte_t pb, nb;
	const xmlbyte_t *p = prev->getBytes();
	xmlbyte_t *d = dest;
	const xmlbyte_t *n = next->getBytes();
	
	// copy the indentical prefix of prev and next
	while((pb = *p) == (nb = *n)) {
		// This would mean next and prev were identical!
		DBXML_ASSERT(pb != 0);
		*d = pb;
		++d; ++p; ++n;
	}
	
	DBXML_ASSERT(nb > pb); // As next should be greater than prev
	
	pb = idbyte2num(pb);
	nb = idbyte2num(nb);
	// Special case where nb equals NS_ID_FIRST,
	// and pb equals 0 - in this case, idbyte2num
	// maps them both onto the number 0.
	while(pb == nb) {
		*(d++) = num2idbyte(nb);
		pb = nextIdbyteAsNum(p);
		++n;
		nb = idbyte2num(*n);
	}
	
	// Pick id so that it is half way between next and prev
	// NOTE: commented-out code will use the "old" algorithm
	// that doesn't consider append vs insertAfter vs insertBefore
	xmlbyte_t avg = avgbyte((int)pb, (int)nb, itype);
	//xmlbyte_t avg = ((int)pb + (int)nb) >> 1;
	*(d++) = num2idbyte(avg);
	while(avg == pb) {
		pb = nextIdbyteAsNum(p);
		avg = avgbyte((int)pb, idbyte2num(NS_ID_LAST) + 1, itype);
		//avg = ((int)pb + idbyte2num(NS_ID_LAST) + 1) >> 1;
		*(d++) = num2idbyte(avg);
	}
	*(d++) = 0; // null terminate
	
	// NOTE: this bypasses setLen*() -- see assertions
	// Use maxlen for comparison, as it was used to determine
	// allocation
	nidLen = (uint32_t) (d - dest);
	if(nidLen > NID_BYTES_SIZE) {
		setLenAlloc(nidLen);
		DBXML_ASSERT(isAlloced());
	} else if (maxlen > NID_BYTES_SIZE) {
		// maintain invariant that smaller NIDs are
		// in nidBytes array.
		memcpy(nidStore.nidBytes, dest, nidLen);
		::free(dest);
	}
#if 0
	std::cout << "Between, where is " << itype <<": ";
	displayNid(std::cout);
	std::cout << std::endl;
#endif
}

void
NsFullNid::setDocRootNid()
{
	freeNid();
	nidStore.nidBytes[0] = NID_ROOT_SIZE;
	nidStore.nidBytes[1] = NS_ROOT_ID;
	nidStore.nidBytes[2] = 0;
	nidLen = (uint32_t)3;
}

void
NsFullNid::setSpecialNid(xmlbyte_t id)
{
	freeNid();
	nidStore.nidBytes[0] = NID_ROOT_SIZE;
	nidStore.nidBytes[1] = id;
	nidStore.nidBytes[2] = 0;
	nidLen = 3;
}

bool
NsFullNid::isDocRootNid() const
{
	return ((getBytes()[1] == NS_ROOT_ID) &&
		(getLen() == 3));
}

bool
NsFullNid::isMetaDataNid() const
{
	return ((getBytes()[1] == NS_ROOT_ID) &&
		(getLen() == 3));
}

//static -- used only for one-byte ids -- metadata, etc
void
NsFullNid::initNid(xmlbyte_t *buf, xmlbyte_t id)
{
	buf[0] = 1;
	buf[1] = id;
	buf[2] = 0;
}

void
NsFullNid::copyNid(const NsFullNid *from)
{
	copyNid(from->getBytes(), from->getLen());
}

void
NsFullNid::copyNid(const xmlbyte_t *fromPtr, uint32_t newLen)
{
	// reuse existing memory if possible
	if (!newLen) {
		freeNid();
		return;
	}
	if (newLen > NID_BYTES_SIZE) {
		if (getLen() < newLen) {
			// cannot reuse
			freeNid();
			nidStore.nidPtr = (xmlbyte_t *)::malloc(newLen);
			if (!nidStore.nidPtr)
				NsUtil::nsThrowException(XmlException::NO_MEMORY_ERROR,
							 "copyNid failed to allocate memory",
							 __FILE__, __LINE__);
			setLenAlloc(newLen);
		} else {
			setLen(newLen); // maintains alloc'd status
		}
		memcpy(nidStore.nidPtr, fromPtr, newLen);
	} else {
		freeNid();
		memcpy(nidStore.nidBytes, fromPtr, newLen);
		setLen(newLen);
	}
	DBXML_ASSERT(getLen() == newLen);
}

// static
void NsNid::displayNid(std::ostream &out, const char *buf, uint32_t len) {
	// find out where the decimal point is
	char n = buf[0] + 0x30;

	++buf;
	--len;
	for(uint32_t i = 0; i < len; ++i, ++buf) {
		if((char)i == n) out << ".";

		unsigned char b = *buf;
		int hn = (b & 0xF0) >> 4;
		int ln = b & 0x0F;
		out << (char)(hn < 10 ? '0' + hn : 'A' + hn - 10);
		out << (char)(ln < 10 ? '0' + ln : 'A' + ln - 10);
	}
}

// NsNidGen methods

NsNidGen::NsNidGen()
	: _idDigit(NID_INITIAL_DIGIT)
{
	memset(_idBuf, 0, NIDBUF_SIZE);
	_idBuf[1] = NS_ROOT_ID;
	_idBuf[0] = NID_ROOT_SIZE;
}

void
NsNidGen::nextId(NsFullNid *id)
{
	if(id != 0) {
#ifdef USE_LARGE_NIDS
		// debugging code to artificially bump Nid size
		// for testing.  Turn it on at top of this file...
		if (_idDigit == 0 && _idBuf[0] == 1 &&
		    _idBuf[1] == 5) {
			// force a large size -- enough to allocate
			_idDigit = 4;
			_idBuf[1] = NS_ID_LAST;
			_idBuf[2] = NS_ID_LAST;
			_idBuf[3] = NS_ID_LAST;
			_idBuf[4] = NS_ID_LAST;
			_idBuf[5] = NS_ID_LAST;
			_idBuf[0] = 5;
		}
#endif
		xmlbyte_t *dest;
		uint32_t len = _idDigit + 3; // incl leading byte and trailing null
		// ensure enough space for the id
		if (len > NID_BYTES_SIZE) {
			dest = id->nidStore.nidPtr =
				(xmlbyte_t *) ::malloc(len);
			if (!dest)
				NsUtil::nsThrowException(XmlException::NO_MEMORY_ERROR,
					"nextId failed to allocate memory",
					__FILE__, __LINE__);
			id->setLenAlloc(len);
		} else {
			dest = id->nidStore.nidBytes;
			id->setLen(len);
		}
		// copy the id
		memcpy(dest, _idBuf, len);
	}
	// Increment the id, for next time
	_nextId(_idBuf, _idDigit); 
}

NsNid::NsNid(const NsFullNid *fullnid) : nid_(0)
{
	if (fullnid)
		nid_ = fullnid->getBytes();
}

// returns:
// 0 if equal
// < 0 if "this" is less than other
// > 0 if "this" is greater than other
// static
int
NsNid::compare(const unsigned char *p1,
	       const unsigned char *p2)
{
	// Assume null termination of node ids
	int res = 0;
	while((res = (int)*p1 - (int)*p2) == 0 &&
	      *p1 != 0) {
		++p1; ++p2;
	}
	return res;
}

bool
NsNid::isDocRootNid() const
{
	return (nid_ && (nid_[1] == NS_ROOT_ID) &&
		(nid_[2] == 0));
}

bool
NsNid::isMetaDataNid() const
{
	return (nid_ && (nid_[1] == NS_METADATA_ID) &&
		(nid_[2] == 0));
}


uint32_t NsNidWrap::set(const unsigned char *nid)
{
	clear();
	uint32_t len = 0;
	if ((const unsigned char *)nid) {
		len = (uint32_t)NsUtil::nsStringLen(nid) + 1;
		nid_ = (const unsigned char*) NsUtil::allocate(len);
		memcpy((void*)nid_, nid, len);
	}
	return len;
}

void NsNidWrap::clear()
{
	if (nid_) NsUtil::deallocate((void*)nid_);
}
