//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002,2008 Oracle.  All rights reserved.
//
// $Id: NsReindexer.cpp,v 1.9 2008/01/22 12:27:55 jpcs Exp $
//

#include "../DbXmlInternal.hpp"
#include "NsReindexer.hpp"
#include "NsUpdate.hpp"
#include "NsDom.hpp"
#include "NsEventReader.hpp"
#include "EventReaderToWriter.hpp"
#include "../Document.hpp"
#include "../UTF8.hpp"

using namespace DbXml;
using namespace std;

// list of elements to index for an element reindex operation
class ElementIndexList {
public:
	ElementIndexList(NsReindexer &reindexer) :top_(0),
		fullGenIndex_(0), startGenIndex_(0), genDocEvents_(false),
		reindexer_(reindexer) {}
	void push(NsNode *node, const std::string &uri, bool fullGeneration,
		  bool startGen);
	void generate(EventWriter &writer);
	void generateStartEvent(NsNode *node, const std::string &uri,
				EventWriter &writer);
	void generateEndEvent(NsNode *node, EventWriter &writer);
	void generateEvents(NsNode *node, EventWriter &writer);
	void setGenDocEvents() { genDocEvents_ = true; }
private:
	vector<NsNodeRef> elements_;
	vector<std::string> elementUris_;
	int top_;
	int fullGenIndex_;
	int startGenIndex_;
	bool genDocEvents_;
	NsReindexer &reindexer_;
};

NsReindexer::NsReindexer(Document &document,
			 OperationContext &oc,
			 int timezone,
			 bool forDelete)
	: Indexer(timezone), key_(timezone), parentKey_(timezone),
	  document_(document)
{
	if (document.getContainerID() == 0)
		return;
	ScopedContainer sc(document.getManager(),
			   document.getContainerID(), true);
	container_ = sc.getContainer();
	DBXML_ASSERT(container_);

	is_.read(container_->getConfigurationDB(), oc.txn(), /*lock=*/false);
	if (forDelete)
		is_.set(Index::INDEXER_DELETE);
	// init remaining Indexer state
	oc_ = &oc;
	initIndexContent(is_, document.getID(), 0, stash_,
			 true, // updateStats -- TBD -- implement option
			 false, // writeNsInfo
			 forDelete);
	dictionary_ = container_->getDictionaryDatabase();
}

void NsReindexer::updateIndexes()
{
	if (!container_)
		return;
	// update the actual index(es) involved
	stash_.updateIndex(*oc_, container_);
	StructuralStatsDatabase *stats = container_->getStructuralStatsDB();
	if (stats) {
		stats->addStats(*oc_, getStatsCache());
		resetStats(); // TBD -- may not be necessary
	}
}

// This code is the same for removal and insertions.
// The difference is in the KeyStash (for delete or not)
void NsReindexer::indexAttributes(NsNodeRef &node)
{
	if (attributesIndexed_ && node->hasAttributes()) {
		for (int i = 0; i < (int)node->numAttrs(); i++) {
			nsAttr_t *attr = node->getAttr(i);
			indexAttribute((const char *)attr->a_name8,
				       node->attrUri(i),
				       node, i);
		}
	}
}

bool NsReindexer::indexElement(NsNodeRef &node)
{
	bool ancestorHasValueIndex = false;
	NsNodeRef newNode(node);
	if ((elementsIndexed_ || updateStats_) && !newNode->isDoc()) {
		const DocID &did = document_.getID();
		DbWrapper &db = *document_.getDocDb();
		ElementIndexList nodes(*this);
		do {
			bool hasValueIndex = false;
			bool hasEdgePresenceIndex = false;
			lookupIndexes(newNode, hasValueIndex, hasEdgePresenceIndex);
			if (hasValueIndex && newNode.get() != node.get())
				ancestorHasValueIndex = true;

			string uri;
			if (newNode->hasUri())
				uri = lookupUri(newNode->uriIndex());
			nodes.push(*newNode, uri, hasValueIndex, hasEdgePresenceIndex);

			newNode = NsUpdate::fetchNode(NsNid(newNode->getParentNid()),
						   did, db, *oc_);
			// tell generator to generate start/end doc events
			// because we're starting at the root.
			if (newNode->isDoc() && hasValueIndex)
				nodes.setGenDocEvents();
		} while(!newNode->isDoc());
		nodes.generate(*this);
	}
	return ancestorHasValueIndex;
}

// reindex if:
//  o there is a container and
//  o it's not a wholedoc/nodes indexed container (which
//  requires full reindexing)
bool NsReindexer::willReindex() const
{
	return (container_ &&
		!(container_->isWholedocContainer() &&
		  container_->nodesIndexed()));
}

// Determine what indexes are present on the node
void NsReindexer::lookupIndexes(NsNodeRef &node, bool &hasValueIndex,
				bool &hasEdgePresenceIndex)
{
	// does this element have a value index (equality or substring)
	const IndexVector *iv = 0;

	string uriname = makeUriName((node->hasUri() ? node->uriIndex() : NS_NOURI),
		(const char *)node->getNameChars());

	if ((iv = is_.getIndexOrDefault(uriname.c_str())) != 0) {
		if (iv->isEnabled(Index::NODE_ELEMENT, Index::NODE_MASK)) {
			if ((iv->isEnabled(Index::XEE, Index::NK_MASK) ||
			     iv->isEnabled(Index::XES, Index::NK_MASK)))
				hasValueIndex = true;
			if (iv->isEnabled(Index::EEP,Index::NK_MASK))
				hasEdgePresenceIndex = true;
		}
	}
}

void NsReindexer::indexAttribute(const char *aname,
				 int auri, NsNodeRef &parent,
				 int attrIndex)
{
	DBXML_ASSERT(container_ && attributesIndexed_);
	// some of the work done here is redundant wrt to Indexer,
	// but it's cheaper here
	const IndexVector *iv = 0;
	string uriname = makeUriName(auri, aname);
	if ((iv = is_.getIndexOrDefault(uriname.c_str())) != 0) {
		if (iv->isEnabled(Index::NODE_ATTRIBUTE, Index::NODE_MASK)) {
			NsNodeIndexNodeInfo ninfo(parent.get());
			indexEntry_.setIndexNodeInfo(&ninfo);
			indexEntry_.setLastDescendant(
				NsNid(parent->getLastDescendantNidOrSelf()));
			key_.reset();
			parentKey_.reset();
			// add value and name id to key
			const char *val = aname;
			while(*(val++)); // get past null to value
			key_.addValue(val, ::strlen(val));
			NameID &id = key_.getID1();
			dictionary_->lookupIDFromName(
				*oc_, uriname.c_str(), id, /* define */ true);
			indexEntry_.setIndex(attrIndex);
			// generate for node indexes
			generateKeys(*iv, Index::NA, Index::PN_MASK, key_,
				     stash_);

			// add parent id, and generate for edge indexes
			string parentUriname =
				makeUriName((parent->hasUri() ? parent->uriIndex() : NS_NOURI),
					    (const char *)parent->getNameChars());
			NameID &parentID = parentKey_.getID1();
			dictionary_->lookupIDFromName(
				*oc_, parentUriname.c_str(), parentID,
				/* define */ false); // name id should exist
			key_.setID2(parentID);
			generateKeys(*iv, Index::EA, Index::PN_MASK, key_,
				     stash_);
		}
	}
}

const char *NsReindexer::lookupUri(int uriIndex)
{
	
	if (uriIndex != NS_NOURI) {
		NameID id(uriIndex);
		return dictionary_->lookupName(*oc_, id);
	}
	return 0;
}

std::string NsReindexer::makeUriName(int uriIndex, const char *lname)
{
	std::string str;
	const char *uri = lookupUri(uriIndex);
	if (uri)
		str = uri;
	str += lname;
	return str;
}

EventReader *NsReindexer::getEventReader(NsNode *node)
{
	NsNid nid(node->getNid());
	return new NsEventReader(
		oc_->txn(),
		document_.getDocDb(),
		dictionary_,
		document_.getID(),
		document_.getContainerID(),
		0, // no flags
		NS_EVENT_BULK_BUFSIZE,
		&nid);
}
	
// ElementIndexList

void ElementIndexList::push(NsNode *node, const std::string &uri, bool fullGeneration,
			    bool startGen)
{
	elements_.push_back(node);
	elementUris_.push_back(uri);
	if (fullGeneration)
		fullGenIndex_ = top_;
	if (startGen || fullGeneration)
		startGenIndex_ = top_;
	++top_;
}

// This is the guts of reindexing elements.
//
// Generate manual start/end events for the parent of
// the top-most indexed element in the tree.  Ignore
// elements from there to the root, since no indexes are affected.
//
// Generate full events for indexed nodes.
//
void ElementIndexList::generate(EventWriter &writer)
{
	DBXML_ASSERT(startGenIndex_ >= fullGenIndex_);

	// if first node is the "root" generate start/end document events
	if (genDocEvents_)
		writer.writeStartDocument(0,0,0);

	// generate simple start events until reaching
	// the full generation index
	int curTop = startGenIndex_ + 1;
	if (curTop == top_) --curTop;

	int current = curTop;
	while (current > fullGenIndex_) {
		// simple generation
		generateStartEvent(elements_[current].get(),
				   elementUris_[current],
				   writer);
		--current;
	}

	// full event generation starting with indexed node
	generateEvents(elements_[fullGenIndex_].get(), writer);

	current = curTop;
	while (current > fullGenIndex_) {
		generateEndEvent(elements_[current].get(), writer);
		--current;
	}

	if (genDocEvents_)
		writer.writeEndDocument();
}

void ElementIndexList::generateStartEvent(NsNode *node,
					  const std::string &uri,
					  EventWriter &writer)
{
	NsNodeIndexNodeInfo ninfo(node);
	writer.writeStartElementWithAttrs(
		(const unsigned char *)node->getNameChars(),
		0, // prefix -- not used
		(const unsigned char *)(uri.c_str()),
		0, 0, // no attrs, attrlist
		&ninfo,
		false);
}

void ElementIndexList::generateEndEvent(NsNode *node, EventWriter &writer)
{
	NsNodeIndexNodeInfo ninfo(node);
	writer.writeEndElementWithNode(0, 0, 0, // no localname, prefix, uri
				       &ninfo);
}

void ElementIndexList::generateEvents(NsNode *node, EventWriter &writer)
{
	AutoDelete<EventReader> reader(reindexer_.getEventReader(node));
	// run the events (do not donate objects to r2w)
	EventReaderToWriter r2w(*reader, writer, false, false);
	r2w.start();
}


