//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002,2008 Oracle.  All rights reserved.
//
// $Id: NsUpdate.cpp,v 1.38 2008/03/04 20:38:14 gmf Exp $
//
#include "../DbXmlInternal.hpp"
#include "NsNid.hpp"
#include "NsUpdate.hpp"
#include "NsFormat.hpp"
#include "NsDom.hpp"
#include "NsEventWriter.hpp"
#include "EventReader.hpp"
#include "EventReaderToWriter.hpp"
#include <xqilla/context/DynamicContext.hpp>
#include "../dataItem/DbXmlConfiguration.hpp"
#include <dbxml/XmlManager.hpp>
#include <dbxml/XmlUpdateContext.hpp>
#include "../Document.hpp"
#include "../UpdateContext.hpp"
#include "../TypeConversions.hpp"
#include "../UTF8.hpp"
#include "NsReindexer.hpp"

using namespace DbXml;
using namespace std;

namespace DbXml {

class NsQName {
public:
	NsQName(ATQNameOrDerived *qname) :
		name_(qname->getName()),
		uri_(qname->getURI()),
		prefix_(qname->getURI() ? qname->getPrefix() : 0) {}
public:
	XMLChToUTF8Null name_;
	XMLChToUTF8Null uri_;
	XMLChToUTF8Null prefix_;
};
}
 
typedef RewriteMap::const_iterator RewriteIter;
typedef UpdateMap::const_iterator MapIter;
typedef pair<MapIter,MapIter> MapRange;
static string makeKey(const NsNid &nid,
		      const DocID &did,
		      const std::string &cname);
static string makeKey(const DbXmlNodeImpl &node);
static void lookupNameIDs(NsQName &qname,
			  const ContainerBase *container,
			  OperationContext &oc,
			  int &uri,
			  int &prefix);


class BetweenNid : public NsNidGen {
public:
	BetweenNid() {}
	virtual ~BetweenNid() {}
	virtual void nextId(NsFullNid *id) {
		NsNid prev(preceding_.nid());
		NsNid next(following_ ? following_.nid() : NsNid());
		// TBD -- GMF -- revisit arg and algorithm
		id->getBetweenNid(&prev, (next ? &next : 0), 3);
		// reset preceding to be new nid
	 	preceding_ = NsNid(id->getBytes());
		if (!first_)
			first_ = id;
	}
	NsNid getFirstNid() {
		DBXML_ASSERT(first_);
		return first_.nid();
	}
public:
	NsFullNidWrap preceding_;
	NsFullNidWrap following_;
	NsFullNidWrap first_;
};


static void getNidRange(BetweenNid &between,
			NsDomNode *prev,
			NsDomNode *next,
			NsDomElement *parent)
{
	// preceding is last desc of prev or the parent
	between.preceding_ = (prev ? prev->getLastDescendantNid() :
			      parent->getNodeId());

	// following is either next or need to search for it...
	// it'll be empty if appending to the end of the document
	if (next) {
		between.following_ = next->getNodeId(); // set following
	} else {
		// search
		// don't put parent into pref because it's stack-allocated
		NsDomNodeRef pref;
		if (parent->hasElemNext()) {
			pref = parent->getElemNext();
			between.following_ = pref->getNodeId();
			return;
		}
		pref = parent->getNsParentNode();
		
		while (pref && !((NsDomElement*)pref.get())->hasElemNext())
			pref = pref->getNsParentNode();
		if (pref) {
			pref = ((NsDomElement*)pref.get())->getElemNext();
			DBXML_ASSERT(pref);
			between.following_ = pref->getNodeId();
		}
	}
}

//
// o next is optional, and if present, may be text...
// o return "fresh" (newly-fetched) nodes
// o a null return is expected if there is no previous element sibling
// o also find out if a "new" previous sibling is text, and if so, return it
//   for text entry moves
//
static NsDomNodeRef findPreviousElem(const NsDomNode *next,
				     NsDomElement &parent,
				     NsDomNodeRef &prevText,
				     DbWrapper &db, OperationContext &oc)
{
	NsDomNodeRef prevRef;
	if (next)
		prevRef = ((NsDomNode*)next)->getNsPrevSibling();

	if (prevRef && prevRef->getNsNodeType() != nsNodeElement)
		prevText = prevRef; // return prev text sibling
	
	while (prevRef && prevRef->getNsNodeType() != nsNodeElement)
		prevRef = prevRef->getNsPrevSibling();
	if (prevRef || next)
		return prevRef;
	// no next -- it's an append or insert into empty element
	return parent.getElemLastChild();
}

//
// o next is optional, and if present, may be text...
// o return "fresh" (newly-fetched) nodes
// o a null return is expected if there is no next element sibling
//
static NsDomNodeRef findNextElem(const NsDomNode *next,
				 const DocID &did,
				 DbWrapper &db, OperationContext &oc)
{
	NsDomNodeRef nextRef;
	if ((next) && (next->getNsNodeType() == nsNodeElement)) {
		// caller has already fetched new node in this case,
		// just duplicate
		return next->duplicate();
	} else if (next) {
		nextRef = (NsDomNode*)next;
		DBXML_ASSERT(nextRef->isTextType());
	}

	while (nextRef && nextRef->getNsNodeType() != nsNodeElement)
		nextRef = nextRef->getNsNextSibling();
	return nextRef;
}

static void initializeTextEntry(nsTextEntry &newText,
				const DbXmlNodeImpl &node)
{
	if (node.getNodeType() == nsNodePinst) {
		XMLChToUTF8 target(node.getPITarget());
		XMLChToUTF8 data(node.getValue());
		NsNode::createPI(&newText.te_text,
				 (const xmlbyte_t *)target.str(),
				 (const xmlbyte_t *)data.str(),
				 false, 0);
		newText.te_type = NS_PINST;
	} else {
		XMLChToUTF8 value(node.getValue());
		bool hasEnt = false;
		NsNode::createText(&newText.te_text,
				   (const xmlbyte_t *)value.str(),
				   value.len(), false, false,
				   &hasEnt);
		switch (node.getNodeType()) {
		case nsNodeText:
			newText.te_type = NS_TEXT;
			if (hasEnt)
				newText.te_type |= NS_ENTITY_CHK;
			break;
		case nsNodeCDATA:
			newText.te_type = NS_CDATA;
			break;
		case nsNodeComment:
			newText.te_type = NS_COMMENT;
			break;
		default:
			DBXML_ASSERT(false);
			break;
		};
	}
}

//
// public methods
//

//
// Insert new element content.
// Called from insert before/after/into/as first/as last.
// o into is modeled as append (insert as last) because that's
//   most efficient for BDB XML
// o append/as last will set neither prev nor next
// o only one of next or prev is allowed
//
// Content is added using NsEventReader plugged into NsEventWriter,
// which (now) allows events to start in the middle of an existing
// document.
//
// Text entries and next/previous plumbing are handled separately.
//
// TBD: split up and/or simplify this function.  It's huge and complex...
//
void NsUpdate::insertElement(const DbXmlNodeImpl &node,
			     const DbXmlNodeImpl &parent,
			     const NsDomNode *next,
			     XmlManager &mgr,
			     Document &document,
			     OperationContext &oc,
			     DynamicContext *context)
{
	XmlUpdateContext uc = mgr.createUpdateContext();
	DbWrapper &db = *(document.getDocDb());
	const DocID &did = document.getID();

	// reindex (this is necessary here in case insertion
	// affects indexes on parent nodes -- mixed content)
	removeElementIndexes(parent, document, oc, /*reindex*/true);

	// fetch fresh parent node for this update and turn it
	// into NsDom for convenience (NsDom will own NsNode and cleanup)
	NsNode *parentNode = fetchNode(parent, db, oc);
	NsDoc tdoc(oc.txn(), &db, document.getDictionaryDB(),
		   did, document.getContainerID(), 0 /* flags -- TBD */);
	NsDomElement parentDom(parentNode, &tdoc);

	// initialize reader and writer
	AutoDelete<EventReader> reader(node.getEventReader(context));
	DBXML_ASSERT(reader.get());
	AutoDelete<NsEventWriter> writer(
		new NsEventWriter(
			&db, document.getDictionaryDB(),
			did,  oc.txn()));
	// refresh next, if present
	if (next)
		const_cast<NsDomNode*>(next)->refreshNode(oc, true);
	// find next and previous elements, if they exist
	NsDomNodeRef previousText; // will hold prev text node, if exists
	NsDomNodeRef previousElem = findPreviousElem(next, parentDom,
						     previousText, db, oc);
	NsDomNodeRef nextElem = findNextElem(next, did, db, oc);
	
	// Calculate node id range for new nodes
	BetweenNid nidGen;
	getNidRange(nidGen, previousElem.get(),
		    nextElem.get(), &parentDom);

	// set writer state for the partial insertion
	writer->setDepth(parentNode->getLevel());
	writer.get()->setNsNidGen(&nidGen);
	writer.get()->setCurrent(parentNode);
	writer.get()->setPrevious(previousElem ?
				  (NsNode*)previousElem->getNsNode() : 0);
	DBXML_ASSERT(!parentNode->getParent());
	// event writer will call NsDocument::completeNode() on
	// parent and previous, so bump the ref count
	parentNode->acquire();
	if (previousElem)
		((NsNode*)previousElem->getNsNode())->acquire();

	// run the events (do not donate objects to r2w)
	EventReaderToWriter r2w(*reader, (NsEventWriter&)*writer, false, false);
	r2w.start();
	writer.get()->end(); // force writing/cleanup


	// now handle "plumbing" -- NID links and text movement

	// if text entries move, they come from either the parent (child text)
	// or from a "next" element.  Moving text as well as setting next/prev
	// on a nextElem will update both the target and nextElem, so amortize
	// the costs of fetch/write
	int startTextIndex = 0;
	NsNode *ownerNode = 0;
	if ((!previousText) &&
	    (!next && parentNode->hasTextChild())) {
		// appending...
		previousText = parentDom.getNsLastChild();
		DBXML_ASSERT(previousText->getNsNodeType() != nsNodeElement);
		startTextIndex = parentNode->getFirstTextChildIndex();
		ownerNode = parentNode;
	}

	if (nextElem || previousText) {
		NsNode *nextElemNode = (nextElem ?
					(NsNode*)nextElem->getNsNode() : 0);
		NsNodeRef target = fetchNode(nidGen.getFirstNid(), did,
					     db, oc);
		// next/prev links
		if (nextElemNode)
			nextElemNode->setNextPrev(target.get());
		// now, text
		nsTextList_t *oldList = 0;
		bool freeTextList = false;
		if (previousText) {
			if (!ownerNode) {
				// owner is either next element or the parent
				ownerNode = (nextElemNode ? nextElemNode : parentNode);
				DBXML_ASSERT(ownerNode);
			}
			int endTextIndex = previousText->getIndex();
			const std::string &cname = document.getContainerName();
			oldList = coalesceTextNodes(ownerNode,
						    *target,
						    startTextIndex,
						    endTextIndex,
						    false,
						    document);
			// remove text from old node
			ownerNode->removeText(startTextIndex, endTextIndex);
			for (int i = startTextIndex; i <= endTextIndex; i++) {
				textRemoved(i, ownerNode->getNid(),
					    did, cname);
			}
			// NOTE: no need to track text inserted -- it's a new node
			// and can't be part of a further update

			// write node if not going to do it below
			if (ownerNode != nextElemNode)
				putNode(ownerNode, db, did, oc);
			freeTextList = true;
		}

		// write nodes
		if (nextElemNode)
			putNode(nextElemNode, db, did, oc);
		putNode(*target, db, did, oc);
		if (freeTextList)
			target->replaceTextList(oldList,
						true /*free oldlist*/);
	}

	// Update last descendants of parent's parent, but only
	// if this was an append
	if (!nextElem) {
		NsDomNodeRef ppRef = parentDom.getElemParent();
		NsFullNid *lastDescNid =
			(NsFullNid *)parentNode->getLastDescendantNid();
		// only update grandparent if parent does not have a 'next'
		// ppRef will be empty if parent id document element
		if (!parentDom.hasElemNext() && ppRef)
			updateLastDescendants(ppRef.get(),
				      lastDescNid,
					      db, did, oc);
	}
	markForUpdate(&document);

}

void NsUpdate::insertText(const DbXmlNodeImpl &node,
			  const DbXmlNodeImpl &parent,
			  const NsDomNode *next,
			  Document &document,
			  OperationContext &oc,
			  DynamicContext *context)
{
	DbWrapper &db = *(document.getDocDb());
	const DocID &did = document.getID();

	// check for empty value
	const XMLCh *tval = node.getValue();
	if (!tval || !*tval)
		return;

	// reindex
	removeElementIndexes(parent, document, oc, /*reindex*/true);

	// refresh next, if present
	if (next)
		const_cast<NsDomNode*>(next)->refreshNode(oc, true);
	NsDomNodeRef nextElem = (next ?
				 findNextElem(next, did, db, oc) :
				 (NsDomNodeRef)0);
	
	// target for new text is either nextElem (if present),
	// which means leading text, or the parent -- child text
	NsNodeRef targetNode = (nextElem ? nextElem->getNsNode() :
				fetchNode(parent, db, oc));

	int index = -1;
	int insertIndex; // track this separately, using original index
	bool asChild = false;
	const std::string &cname = document.getContainerName();
	if (next && next->isTextType()) {
		insertIndex = next->getIndex();
		index = getTextIndex(insertIndex,
				     next->getNodeId(),
				     did,
				     cname);
		if (((NsDomText*)next)->isChildText())
			asChild = true;
	} else if (next) {
		// last leading text...
		insertIndex = index = targetNode->getNumLeadingText();
		
	} else { // trailing child text
		insertIndex = index = targetNode->getNumText();
		asChild = true;
	}

	// insert new text at "index" in target node
	nsTextEntry_t newText;

	initializeTextEntry(newText, node);
	
	nsTextList_t *oldTextList = insertInTextList(&newText,
						     index,
						     *targetNode,
						     false, asChild);
	textInserted(insertIndex, targetNode->getNid(),
		     did, cname);
	
	// Re-write the node, using new text list
	putNode(*targetNode, db, did, oc);

	// note for possible text node coalesce
	if (targetNode->getNumLeadingText() >= 2 ||
	    targetNode->getNumChildText() >= 2)
		markElement(textCoalesce_, targetNode->getNid(), document, false);

	// Put things back the way they were for cleanup
	(void) targetNode->replaceTextList(oldTextList, true);
	markForUpdate(&document);
}

void NsUpdate::insertAttributes(
	const std::vector<const DbXmlNodeImpl*> &attributes,
	const DbXmlNodeImpl &parent,
	Document &document,
	OperationContext &oc,
	DynamicContext *context)
{
	DbWrapper &db = *(document.getDocDb());
	const DocID &did = document.getID();
	NsDocument *doc =
		const_cast<Document*>(parent.getDocument())->getNsDocument();
	
	// fetch fresh node
	NsNodeRef parentNode = fetchNode(parent, db, oc);

	// reindex
	removeAttributeIndexes(parentNode, document, oc);
	
	// allocate new list, copy parent's existing attrs, and
	// add new ones
	int numNewAttrs = (int)attributes.size();
	nsAttrList_t *oldList =
		parentNode->copyAttrList(numNewAttrs);

	for (int i = 0; i < numNewAttrs; i++) {
		const DbXmlNodeImpl *node = attributes[i];
		(void) parentNode->addAttr(doc,
					   node->getPrefix(),
					   node->getUri(),
					   node->getLocalName(),
					   node->getValue(),
					   true);
	}

	// Re-write the node, using new attribute list
	putNode(*parentNode, db, did, oc);

	// Put things back the way they were for cleanup
	(void) parentNode->replaceAttrList(oldList, true);
	markForUpdate(&document);
}


void NsUpdate::removeDocument(const DbXmlNodeImpl &node,
			      Container &container,
			      XmlManager &mgr,
			      OperationContext &oc)
{
	// Remove entire document
	XmlUpdateContext uc = mgr.createUpdateContext();
	container.deleteDocumentInternal(
		oc.txn(),
		(Document&)*node.getDocument(),
		uc);
}

void NsUpdate::removeAttribute(const DbXmlNodeImpl &node,
			       Document &document,
			       OperationContext &oc,
			       DynamicContext *context)
{
	// fetch fresh owner node
	DbWrapper *db = document.getDocDb();
	NsNodeRef parentNode = fetchNode(node, *db, oc);

	//reindex
	removeAttributeIndexes(parentNode, document, oc);
	
	// Find the index of the attribute to remove
	int oldindex = getAttributeIndex(node);

	// Update it
	parentNode->removeAttr(oldindex);

	// Note removal
	attributeRemoved(node);
	
	// Re-write the node
	putNode(*parentNode, *db, node.getDocID(), oc);
	markForUpdate(&document);

}

void NsUpdate::removeElement(const DbXmlNodeImpl &node,
			     Document &document,
			     OperationContext &oc,
			     DynamicContext *context)
{
	bool nextModified = false;
	bool prevModified = false;

	// reindex
	removeElementIndexes(node, document, oc, /*reindex*/false);

	// fetch fresh node
	DbWrapper &db = *(document.getDocDb());
	const DocID &did = document.getID();
	NsNodeRef newNode = fetchNode(node, db, oc);
	// Create NsDomElement for navigational methods
	NsDomElement element(newNode.get(),
			     const_cast<Document*>(node.getDocument())->
			     getNsDocument());
	// Fetch nodes that may be modified (to start with)
	NsDomNodeRef parentRef = element.getElemParent();
	NsDomNodeRef nextRef = element.getElemNext();
	NsDomNodeRef prevRef = element.getElemPrev();

	NsNode *nextNode = 0;
	NsNode *prevNode = 0;
	NsNode *parentNode = 0;
	nsTextList_t *oldList = 0;
	bool freeTextList = false;
	NsNode *toNode = 0;
	if (nextRef) {
		nextModified = true;
		nextNode = (NsNode *)nextRef->getNsNode();
		if (prevRef) {
			prevModified = true;
			prevNode = (NsNode*)prevRef->getNsNode();
			nextNode->setNextPrev(prevNode);
		} else
			nextNode->clearPrev();
		// move any leading text
		if (newNode->hasLeadingText()) {
			oldList = coalesceTextNodes(*newNode,
						    nextNode,
						    -1, -1, // all leading text
						    false/* toChild */,
						    document);
			toNode = nextNode;
			freeTextList = true;
		}
	} else {
		parentNode = (NsNode*)parentRef->getNsNode();
		const NsFullNid *lastDescNid = 0;
		// this was last child element...
		if (prevRef) {
			prevModified = true;
			prevNode = (NsNode*)prevRef->getNsNode();
			prevNode->clearNext();
			// new last descendant is that of the previous node
			lastDescNid = prevNode->getLastDescendantNidOrSelf();
		} else {
			// no other children of parent; it's now a leaf
			lastDescNid = parentNode->getFullNid();
		}
		// prevNode of 0 clears parent's lastChild
		parentNode->setLastChild(prevNode);
		if (newNode->hasLeadingText()) {
			oldList = coalesceTextNodes(
				*newNode,
				parentNode,
				-1, -1, // all leading text
				true /* toChild */,
				document);
			toNode = parentNode;
			freeTextList = true;
		}
		// this updates the parent and all of its parents to the root
		updateLastDescendants(parentRef.get(),
				      lastDescNid,
				      db, did, oc);
		// doc node is not modified by updateLastDescendants;
		// handle case where root element is deleted
		if (parentNode->isDoc() && !parentNode->hasChildElem())
			putNode(parentNode, db, did, oc);
	}
	// Remove node and subtree (combine w/indexing?)
	deleteTree(*newNode, db, did, oc);

	// Write prev/next if present and modified
	if (prevModified)
		putNode(prevNode, db, did, oc);
	if (nextModified)
		putNode(nextNode, db, did, oc);
	if (freeTextList) {
		DBXML_ASSERT(toNode);
		toNode->replaceTextList(oldList,
					true /*free oldlist*/);
	}
	markForUpdate(&document);
}

void NsUpdate::removeText(const DbXmlNodeImpl &node,
			  Document &document,
			  OperationContext &oc,
			  DynamicContext *context)
{
	// reindex
	removeElementIndexes(node, document, oc, /*reindex*/true);

	// fetch fresh owner node
	DbWrapper *db = document.getDocDb();
	NsNodeRef newNode = fetchNode(node, *db, oc);

	// Find the index of the text to remove
	int oldindex = getTextIndex(node);

	// Update it
	newNode->removeText(oldindex, oldindex);

	// Note removal
	textRemoved(node, oldindex);

	// Re-write the node
	putNode(*newNode, *db, node.getDocID(), oc);

	// note for possible text node coalesce
	if (newNode->getNumLeadingText() >= 2 ||
	    newNode->getNumChildText() >= 2)
		markElement(textCoalesce_, newNode->getNid(), document, false);
	
	markForUpdate(&document);
}

void
NsUpdate::renameElement(const DbXmlNodeImpl &node,
			ATQNameOrDerived *qname,
			Document &document,
			OperationContext &oc,
			DynamicContext *context)
{
	// reindex
	removeElementIndexes(node, document, oc, /*reindex*/true);

	// fetch fresh node
	DbWrapper *db = document.getDocDb();
	NsNodeRef newNode = fetchNode(node, *db, oc);
	NsQName nqname(qname);

	// lookup uri and prefix if provided
	int uri = NS_NOURI;
	int prefix = NS_NOPREFIX;
	if (nqname.uri_.len())
		lookupNameIDs(nqname, node.getContainer(), oc,
			      uri, prefix);

	// Rename the element
	newNode->renameElement(nqname.name_.str(),
			       nqname.name_.len(),
			       uri, prefix);
	// Re-write the node
	putNode(*newNode, *db, node.getDocID(), oc);
	markForUpdate(&document);
}

//
// private instance methods
//
int NsUpdate::getAttributeIndex(const DbXmlNodeImpl &node) const
{
	int index = node.getIndex();
	int offset = 0;
	string key = makeKey(node);

	MapRange it = attrMap_.equal_range(key);
	for (MapIter i = it.first; i != it.second; i++) {
		if (i->second < index)
			offset++;
	}
	return index - offset;
}

// TBD: GMF -- figure out a better algorithm or data structure
// for handling inserts and deletes
int NsUpdate::getTextIndex(const DbXmlNodeImpl &node) const
{
	return getTextIndex(makeKey(node), node.getIndex());
}

int NsUpdate::getTextIndex(int index, const NsNid &nid,
			   const DocID &did, const std::string &cname)
{
	return getTextIndex(makeKey(nid, did, cname),
			    index);
}

int NsUpdate::getTextIndex(const std::string &key, int index) const
{
	int offset = 0;
	// subtract one for each deleted entry
	MapRange itdel = textDeleteMap_.equal_range(key);
        MapIter i;
	for (i = itdel.first; i != itdel.second; i++) {
		if (i->second < index)
			offset++;
	}

	// add one for each inserted entry
	MapRange itins = textInsertMap_.equal_range(key);
	for (i = itins.first; i != itins.second; i++) {
		if (i->second <= index)
			offset--;
	}
	return index - offset;
}

void NsUpdate::attributeRemoved(const DbXmlNodeImpl &node)
{
	string key = makeKey(node);
	attrMap_.insert(make_pair<const std::string, int>(key,node.getIndex()));
}

void NsUpdate::textRemoved(const DbXmlNodeImpl &node, int index)
{
	string key = makeKey(node);
	textDeleteMap_.insert(make_pair<const std::string, int>(key,index));
}

void NsUpdate::textRemoved(int index, const NsNid &nid,
			   const DocID &did,
			   const std::string &cname)
{
	string key = makeKey(nid, did, cname);
	textDeleteMap_.insert(make_pair<const std::string, int>(key,index));
}

void NsUpdate::textInserted(int index, const DbXmlNodeImpl &node)
{
	string key = makeKey(node);
	textInsertMap_.insert(make_pair<const std::string, int>(key,index));
}

void NsUpdate::textInserted(int index, const NsNid &nid,
			    const DocID &did,
			    const std::string &cname)
{
	string key = makeKey(nid, did, cname);
	textInsertMap_.insert(make_pair<const std::string, int>(key,index));
}

//
// Insert a text entry into the text list for the
// "to" node at the specified index.  Options:
//  o replace -- replace existing entry
//  o asChild -- this is a (new) child entry (ignored
//    if replace is true)
//
nsTextList_t *NsUpdate::insertInTextList(nsTextEntry_t *text,
					 int index,
					 NsNode *to,
					 bool replace,
					 bool asChild)
{
	nsTextEntry_t *toEntries = 0;
	int numToChild = 0;
	int toIndex = 0;
	if (to->hasText()) {
		toEntries = to->getTextEntry(0);
		numToChild = to->getNumChildText();
	}

	int numText = to->getNumText();
	int numInserted = 0;
	if (!replace)
		++numText;
	nsTextList_t *newTextList = NsNode::createTextList(numText);
	try { // clean up on failure
		while (numInserted < numText) {
			if (numInserted == index) {
				// set NS_DONTDELETE to avoid copy.
				// new text is dynamically allocated, and the
				// property of the new text list
				NsNode::addText(
					newTextList,
					text->te_text.t_chars,
					text->te_text.t_len,
					text->te_type | NS_DONTDELETE,
					false /*not donated*/);
				// clear NS_DONTDELETE to ensure deallocation
				newTextList->tl_text[numInserted].te_type = text->te_type;
				numInserted++;
				continue;
			}
			if (toEntries) {
				if ((toIndex != index) || !replace) {
					NsNode::addText(
						newTextList,
						toEntries[toIndex].te_text.t_chars,
						toEntries[toIndex].te_text.t_len,
						toEntries[toIndex].te_type|NS_DONTDELETE,
						false /* ignored in this case */);
					numInserted++;
				}
				++toIndex;
			}
		}
		DBXML_ASSERT((int)newTextList->tl_ntext == numText);
	} catch (XmlException &xe) {
		NsNode::freeTextList(newTextList);
	}
	//
	// resulting from state change:
	// nchild and related flags
        //

	if (asChild && !replace)
		numToChild++;
	if (numToChild) {
		newTextList->tl_nchild = numToChild;
		to->setFlag(NS_HASTEXTCHILD);
	}
	to->setFlag(NS_HASTEXT);
	nsTextList_t *oldList = to->replaceTextList(newTextList);
	return oldList;
}

//
// Helper function to "move" text from one node to another.
// o if toChild, nodes are move to child text
// o if startIndex or endIndex is other than -1, use the value; otherwise
//   use all leading text
//
// 1. create a temporary text list with values from both
// "from" and "to"
// 2. Replace "to's" list with the new one, returning the old
// list to be replaced by the caller when the node has been
// written.
// 
nsTextList_t * NsUpdate::coalesceTextNodes(NsNode *from,
					   NsNode *to,
					   int startIndex,
					   int endIndex,
					   bool toChild,
					   Document &document)
{
	DBXML_ASSERT(from && to);
	if (startIndex == -1)
		startIndex = 0;
	if (endIndex == -1)
		endIndex = from->getNumLeadingText() - 1;
	uint32_t listIndex = 0;
	nsTextEntry_t *toEntries = 0;
	int numToChild = 0;
	if (to->hasText()) {
		toEntries = to->getTextEntry(0);
		numToChild = to->getNumChildText();
		if (toChild)
			listIndex = to->getFirstTextChildIndex();
	}

	int numFrom = endIndex - startIndex + 1;
	int numText = numFrom + to->getNumText();
	// create new text list to hold all the entries
	nsTextList_t *newTextList = NsNode::createTextList(numText);
	int toIndex = 0;
	int lastType = -1;
	try { // clean up on failure
		for (int i = 0; i < numText; i++) {
			if (i == (int) listIndex) {
				// add new entries
				nsTextEntry_t *fromEntries =
					from->getTextEntry(startIndex);
				for (int j = 0; j < numFrom; j++) {
					NsNode::addText(
						newTextList,
						fromEntries[j].te_text.t_chars,
						fromEntries[j].te_text.t_len,
						fromEntries[j].te_type,
						false /*not donated*/);
					textInserted(i + j,
						     to->getNid(),
						     document.getID(),
						     document.getContainerName());
					lastType = fromEntries[j].te_type;
				}
				i += numFrom - 1;
				continue;
			}
			if (toEntries) {
				// first node may be a candidate for coalescing
				if (nsTextType(lastType) == NS_TEXT &&
				    nsTextType(toEntries[toIndex].te_type) == NS_TEXT &&
				    (toChild || (toIndex != to->getFirstTextChildIndex()))) {
					// that's enough to warrant checking later
					markElement(textCoalesce_, to->getNid(), document, false);
				}
				lastType = -1;
				NsNode::addText(
					newTextList,
					toEntries[toIndex].te_text.t_chars,
					toEntries[toIndex].te_text.t_len,
					toEntries[toIndex].te_type|NS_DONTDELETE,
					false /* ignored in this case */);
				++toIndex;
			}
		}
		DBXML_ASSERT((int)newTextList->tl_ntext == numText);
	} catch (XmlException &xe) {
		NsNode::freeTextList(newTextList);
	}
	//
	// resulting from state change:
        //
	newTextList->tl_nchild = numToChild;
	if (toChild) {
		newTextList->tl_nchild += numFrom;
		to->setFlag(NS_HASTEXTCHILD);
	}
	to->setFlag(NS_HASTEXT);
	nsTextList_t *oldList = to->replaceTextList(newTextList);
	return oldList;
}

bool NsUpdate::coalesceEntry(nsTextList_t *list, nsTextEntry_t *entries, int &lastType, int index)
{
	bool ret = false;
	if (nsTextType(lastType) == NS_TEXT &&
	    nsTextType(entries[index].te_type) == NS_TEXT) {
		DBXML_ASSERT(index > 0);
		NsNode::coalesceText(list, entries[index].te_text.t_chars,
				     entries[index].te_text.t_len, false);
		// coalescing will always eliminate the NS_DONTDELETE flag
		list->tl_text[list->tl_ntext-1].te_type =
			(entries[index].te_type & ~NS_DONTDELETE);
		ret = true;
	} else {
		NsNode::addText(list, entries[index].te_text.t_chars,
				entries[index].te_text.t_len,
				entries[index].te_type, false /* not donated */);
	}
	lastType = entries[index].te_type;
	return ret;
}

void NsUpdate::doCoalesceText(NsNode *node, Document &doc, OperationContext &oc)
{
	int32_t numLeading = node->getNumLeadingText();
	int32_t numChild = node->getNumChildText();
	DBXML_ASSERT(numLeading >= 2 || numChild >= 2);

	nsTextList_t *newList = NsNode::createTextList(node->getNumText());
	nsTextEntry_t *entries = node->getTextEntry(0);
	
	int i, lastType = -1;
	for (i = 0; i < numLeading; i++)
		(void) coalesceEntry(newList, entries, lastType, i);
	if (numChild) {
		lastType = -1;
		for (i = node->getFirstTextChildIndex(); i < node->getNumText(); i++)
			if (coalesceEntry(newList, entries, lastType, i))
				--numChild; // one fewer child text nodes
	}
	newList->tl_nchild = numChild;
	// replace the list, put the node, put things back for releasing
	nsTextList_t *oldList = node->replaceTextList(newList);
	putNode(node, *doc.getDocDb(), doc.getID(), oc);
	node->replaceTextList(oldList, true /*free oldlist*/);
}

void NsUpdate::coalesceText(OperationContext &oc)
{
	for (NidMap::iterator i = textCoalesce_.begin();
	     i != textCoalesce_.end(); i++) {
		Document *doc = i->second.getDocument();
		try {
			NsNodeRef node =
				fetchNode(NsNid(i->second.getNid()),
					  doc->getID(),
					  *doc->getDocDb(),
					  oc);
			if (node->canCoalesceText())
				doCoalesceText(*node, *doc, oc);
		} catch (XmlException &xe) {
			// ignore DB_NOTFOUND in this path --
			// nodes may have been deleted
			if ((xe.getExceptionCode() == XmlException::DATABASE_ERROR) &&
			    (xe.getDbErrno() == DB_NOTFOUND))
				continue;
			throw;
		}
	}
}

void NsUpdate::completeUpdate(XmlManager &mgr,
			      OperationContext &oc,
			      DynamicContext *context)
{
	coalesceText(oc); // coalesce text nodes
	
	reindex(oc); // reindex elements
	
	for (RewriteIter i = documentUpdates_.begin();
	     i != documentUpdates_.end(); i++) {
		// Update document content without indexing
		Document *doc = i->second;
		ScopedContainer sc(mgr, doc->getContainerID(), true);
		Container *cont = sc.getContainer();
		DBXML_ASSERT(cont);
		// this path can occur if nodes were created via Sequential Scan
		// iterators that did not materialize a Document object
		if (doc->getDefinitiveContent() != Document::DOM) {
			DBXML_ASSERT(doc->getDocDb());
			doc->setContentAsNsDom(doc->getID(),doc->getCacheDatabase());
			doc->setContentModified(true);
		}
					       
		// for wholedoc docs with node indexes, need to totally reindex
		if (cont->isWholedocContainer() && cont->nodesIndexed()) {
			XmlUpdateContext uc = mgr.createUpdateContext();
			cont->updateDocumentInternal(oc.txn(), *doc, uc);
		} else {
			DocumentDatabase *db = cont->getDocumentDB();
			doc->getID().setDbtFromThis(oc.key());
			Dbt *dbt = (Dbt*)doc->getContentAsDbt();
			db->getContentDatabase().put(oc.txn(), &oc.key(),
						     dbt, 0);
		}
	}
}

void NsUpdate::markForUpdate(Document *doc)
{
	// only mark if (1) wholedoc and (2) there is a container
	// isWholedoc() handles both.  Transient documents need to
	// be marked as modified and dbt content cleared
	if (doc->getContainerID() == 0 || doc->isWholedoc()) {
		doc->setContentModified(true);
		doc->resetContentAsDbt();
		if (doc->isWholedoc()) {
			// mark for content to be updated
			// use docid and container name as key
			string key = doc->getID().asString();
			key += doc->getContainerName();
			documentUpdates_.insert(make_pair(key,doc));
		}
	}
}

//
// Reindex-related functions
//
void NsUpdate::removeAttributeIndexes(NsNodeRef &parentNode,
				      Document &document,
				      OperationContext &oc)
{
	// remove all attribute indexes for the referenced node (indexes
	// on the node itself remain intact)
	// Node is always reindexed for attributes later
	NsReindexer reindexer(document, oc,
			      document.getManager().getImplicitTimezone(),
			      /* forDelete */true);
	if (reindexer.willReindex()) {
		string key = makeKey(parentNode->getNid(), document.getID(),
				     document.getContainerName());
		// avoid doing this twice
		if (!indexesRemoved(key, true)) {
			reindexer.indexAttributes(parentNode);
			reindexer.updateIndexes();
			markElement(elements_, key, parentNode->getNid(),
				    document, /* attributes */true);
		}
	}
}

void NsUpdate::removeElementIndexes(const DbXmlNodeImpl &node,
				    Document &document,
				    OperationContext &oc, bool reindex)
{
	// do not do anything for document nodes
	if (node.getNodeType() == nsNodeDocument)
		return;
	// remove indexes related to element.  If reindex is true, mark
	// it for reindexing (for update) later
	NsReindexer reindexer(document, oc,
			      document.getManager().getImplicitTimezone(),
			      /* forDelete */true);
	if (reindexer.willReindex()) {
		string key = makeKey(node.getNodeID(), document.getID(),
				     document.getContainerName());
		// avoid doing this twice
		if (!indexesRemoved(key, false)) {
			NsNodeRef nodeRef = node.getNsDomNode()->getNsNode();
			NsNid reindexNid = node.getNodeID();
			if (reindexer.indexElement(nodeRef)) {
				reindex = true; // force reindex of parent
				reindexNid = NsNid(nodeRef->getParentNid());
			}
			reindexer.updateIndexes();
			if (reindex)
				markElement(elements_, key, reindexNid,
					    document, /* attributes */false);
		}
	}
}

void NsUpdate::markElement(NidMap &nidmap, const NsNid &nid, Document &doc,
			   bool attributes)
{
	markElement(nidmap, makeKey(nid, doc.getID(),
                                    doc.getContainerName()),
		    nid, doc,
		    attributes);
}

void NsUpdate::markElement(NidMap &nidmap, const std::string &key,
			   const NsNid &nid, Document &doc,
			   bool attributes)
{
	NidMap::iterator i = nidmap.find(key);
	if (i != nidmap.end()) {
		i->second.setAttributes(attributes);
	} else {
		nidmap.insert(make_pair(key,
				     NidMarker(&doc, nid, attributes)));
	}
}

// is the element specified by key already on the list?
bool NsUpdate::indexesRemoved(const std::string &key,
			      bool attributes) const
{
	NidMap::const_iterator i = elements_.find(key);
	if (i == elements_.end())
		return false;
	// if the previous removal was only attributes
	// and this one is not, it must proceed
	if (i->second.attributesOnly() && !attributes)
		return false;
	return true;
}

void NsUpdate::reindex(OperationContext &oc)
{
	for (NidMap::iterator i = elements_.begin();
	     i != elements_.end(); i++) {
		Document *doc = i->second.getDocument();
		NsReindexer reindexer(
			*doc, oc, doc->getManager().getImplicitTimezone(),
			/* forDelete */false);
		if (reindexer.willReindex()) {
			try {
				// TBD: can probably avoid this fetch
				// in some instances (e.g. attributesOnly
				// and attrs aren't indexed)
				NsNodeRef node =
					fetchNode(NsNid(i->second.getNid()),
						  doc->getID(),
						  *doc->getDocDb(),
						  oc);
				
				if (i->second.attributesOnly()) {
					reindexer.indexAttributes(node);
				} else {
					reindexer.indexElement(node);
				}
				reindexer.updateIndexes();
			} catch (XmlException &xe) {
				// ignore DB_NOTFOUND in this path --
				// nodes may have been deleted
				if ((xe.getExceptionCode() == XmlException::DATABASE_ERROR) &&
				    (xe.getDbErrno() == DB_NOTFOUND))
					continue;
				throw;
			}
		}
	}
}
	
//
// utility functions
//

//static
NsNode *NsUpdate::fetchNode(const DbXmlNodeImpl &node,
			    DbWrapper &db,
			    OperationContext &oc)
{
	return fetchNode(node.getNodeID(), node.getDocID(), db, oc);
}

//static
NsNode *NsUpdate::fetchNode(const NsNid &nid,
			    const DocID &did,
			    DbWrapper &db,
			    OperationContext &oc)
{
	return NsFormat::fetchNode(nid, did, db, oc, true);
}

// static
void NsUpdate::putNode(const NsNode *node,
		       DbWrapper &db,
		       const DocID &did,
		       OperationContext &oc)
{
	int err = NsFormat::putNodeRecord(db, oc, did, node,
					  false /* not adding */);
	if (err)
		throw XmlException(err);
}

// static
void NsUpdate::deleteTree(NsNode *node,
			  DbWrapper &db,
			  const DocID &did,
			  OperationContext &oc)
{
	// delete the range of records from current to
	// last descendant
	Cursor cursor(db, oc.txn(), CURSOR_WRITE, 0);
	if (cursor.error() != 0)
		throw XmlException(cursor.error());
	DbtOut data;
	data.set_flags(DB_DBT_PARTIAL); // only need keys

	// Find last node (maybe the first one)
	DbtOut tmp;
	NsNid lastNid(node->getLastDescendantNidOrSelf());
	NsFormat::marshalNodeKey(did, lastNid, tmp);
	void *lastData = tmp.get_data();
	u_int32_t keySize = tmp.get_size();

	// First nid (this one)
	NsFormat::marshalNodeKey(did, node->getNid(), oc.key());
	int err = cursor.get(&oc.key(), &data, DB_SET_RANGE);

	while (err == 0) {
		try {
			err = cursor.del(0);
			if(Log::isLogEnabled(Log::C_NODESTORE, Log::L_DEBUG)) {
				unsigned char *ptr = (unsigned char *)oc.key().get_data();
				DocID tdid;
				ptr += tdid.unmarshal(ptr);
				NsNid nid(ptr);
				NsFormat::logNodeOperation(db, did, nid,
							   0, "deleted", err);
			}
			if (err == 0 &&
			    oc.key().get_size() == keySize &&
			    memcmp(oc.key().get_data(), lastData, keySize) == 0)
				break;
			if (err == 0)
				err = cursor.get(&oc.key(), &data, DB_NEXT);
		} catch (...) {
			cursor.close();
			throw;
		}
	}
	cursor.close();
}


// NOTE: the first node in the call *must* require modification.
// After that, it depends on if the last modified element has a 'next'
// static
void NsUpdate::updateLastDescendants(NsDomNode *node,
				     const NsFullNid *nid,
				     DbWrapper &db,
				     const DocID &did,
				     OperationContext &oc)
{
	// stop when either (1) at document node or (2) current
	// node has a next element
	NsDomNodeRef nodeRef(node);
	while (nodeRef) {
		NsNode *nsNode = (NsNode*)nodeRef->getNsNode();
		if (nsNode->isDoc())
			break;
		nsNode->setLastDescendantNid(nid);
		putNode(nsNode, db, did, oc);
		if (nsNode->hasNext())
			break;
		nodeRef = nodeRef->getNsParentNode();
	}
}

//
// Note on insert/delete of attributes and text
//
// Insertion and deletion of attribute and text entries affects both
// the "real" index databases and the index values associated with a given
// DbXmlNodeImpl object.  Those index values indicate the index of the
// attribute or text within the owning node.
//
// As attribute or text are added or deleted, the indexes of any
// affected objects need to be adjusted.  In order to do that,
// a map of node->index added/removed is maintained and used.
//

// TBD: GMF -- more efficient way to make a key
// It needs to encapsulate:
//  node ID, doc ID and container name
//
static string makeKey(const NsNid &nid,
		      const DocID &did,
		      const std::string &cname)
{
	string ret = (const char *)nid.getBytes();
	ret += did.asString();
	ret += cname;
	return ret;
}

static string makeKey(const DbXmlNodeImpl &node)
{
	return makeKey(node.getNodeID(),
		       node.getDocID(),
		       node.getContainer()->getName());
}

static void lookupNameIDs(NsQName &qname,
			  const ContainerBase *container,
			  OperationContext &oc,
			  int &uri,
			  int &prefix)
{
	NameID uriID;
	const_cast<ContainerBase*>(container)->lookupID(oc, qname.uri_.str(),
			    qname.uri_.len(),
			    uriID);
	uri = (int)uriID.raw();
	if (qname.prefix_.len()) {
		NameID prefixID;
		const_cast<ContainerBase*>(container)->
		lookupID(oc, qname.prefix_.str(),
			 qname.prefix_.len(),
			 prefixID);
		prefix = (int) prefixID.raw();
	}
}
