//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002,2008 Oracle.  All rights reserved.
//
// $Id: NsUpdate.hpp,v 1.16 2008/03/18 20:52:09 gmf Exp $
//

#ifndef __DBXMLNSUPDATE_HPP
#define __DBXMLNSUPDATE_HPP

#include <string>
#include <map>
#include <vector>
#include "NsNid.hpp"

/*
 * NsUpdate.hpp
 *
 * Class and utility functions that implement node storage format
 * updating functions used to modify existing nodes.
 *
 * This could properly be considered part of NsFormat, but is split
 * into a separate class/file for size and complexity reasons.
 * See NsFormat.cpp for format details.
 */

class DynamicContext;
class ATQNameOrDerived;
class BetweenNid;

namespace DbXml
{

class OperationContext;
class DbWrapper;
class Document;
class Container;
class DbXmlNodeImpl;
class DocID;
class NsNode;
class NsFullNid;
class NsNid;
class NsDomNode;
class XmlManager;
class NsQName;
struct nsTextList;
struct nsTextEntry;
class NsNodeRef;

// NidMarker is used to track nodes that need
// reindexing at the end of an update operation
class NidMarker {
public:
	NidMarker() {}
	NidMarker(Document *doc, const NsNid &nid,
		  bool attrsOnly)
		: doc_(doc), attrsOnly_(attrsOnly),
		  nid_(nid) {}
	NidMarker(const NidMarker &other)
		: doc_(other.doc_), attrsOnly_(other.attrsOnly_),
		  nid_(const_cast<NidMarker&>(other).nid_.get()) {}
	Document *getDocument() {return doc_;}
	NsFullNid *getNid() { return nid_.get(); }
	void setAttributes(bool attrs) {
		if (!attrs && attrsOnly_)
			attrsOnly_ = false;
	}
	bool attributesOnly() const { return attrsOnly_; }
private:
	Document *doc_;
	bool attrsOnly_;
	NsFullNidWrap nid_;
};

typedef std::multimap<std::string, int> UpdateMap;
typedef std::map<std::string, Document*> RewriteMap;
typedef std::map<std::string, NidMarker> NidMap;

class NsUpdate {
public:
	NsUpdate() {}
	~NsUpdate() {}
	// insertion
	void insertElement(const DbXmlNodeImpl &node,
			   const DbXmlNodeImpl &parent,
			   const NsDomNode *next,
			   XmlManager &mgr,
			   Document &document,
			   OperationContext &oc,
			   DynamicContext *context);

	void insertText(const DbXmlNodeImpl &node,
			const DbXmlNodeImpl &parent,
			const NsDomNode *next,
			Document &document,
			OperationContext &oc,
			DynamicContext *context);

	void insertAttributes(
		const std::vector<const DbXmlNodeImpl*> &attributes,
		const DbXmlNodeImpl &parent,
		Document &document,
		OperationContext &oc,
		DynamicContext *context);

	// removal
	void removeDocument(const DbXmlNodeImpl &node,
			    Container &container,
			    XmlManager &mgr,
			    OperationContext &oc);
	void removeAttribute(const DbXmlNodeImpl &node,
			     Document &document,
			     OperationContext &oc,
			     DynamicContext *context);
	void removeElement(const DbXmlNodeImpl &node,
			   Document &document,
			   OperationContext &oc,
			   DynamicContext *context);
	void removeText(const DbXmlNodeImpl &node,
			Document &document,
			OperationContext &oc,
			DynamicContext *context);
	// rename
	void renameElement(const DbXmlNodeImpl &node,
			   ATQNameOrDerived *qname,
			   Document &document,
			   OperationContext &oc,
			   DynamicContext *context);
	// used to rewrite modified wholedoc documents
	void completeUpdate(XmlManager &mgr,
			    OperationContext &oc,
			    DynamicContext *context);
private:
	// internal utility functions
	int getAttributeIndex(const DbXmlNodeImpl &node) const;
	int getTextIndex(const DbXmlNodeImpl &node) const;
	int getTextIndex(int index, const NsNid &nid,
			 const DocID &did, const std::string &cname);
	int getTextIndex(const std::string &key, int index) const;
	void attributeRemoved(const DbXmlNodeImpl &node);
	void textRemoved(const DbXmlNodeImpl &node, int index);
	void textRemoved(int index, const NsNid &nid,
			 const DocID &did,
			 const std::string &cname);
	void textInserted(int index, const DbXmlNodeImpl &node);
	void textInserted(int index, const NsNid &nid,
			  const DocID &did,
			  const std::string &cname);
	struct nsTextList *coalesceTextNodes(NsNode *from,
					     NsNode *to,
					     int startIndex,
					     int endIndex,
					     bool toChild,
					     Document &doc);
	struct nsTextList *insertInTextList(struct nsTextEntry *text,
					    int index,
					    NsNode *to,
					    bool replace,
					    bool asChild);
	// reindexing
	void removeAttributeIndexes(NsNodeRef &parentNode, Document &doc,
				    OperationContext &oc);
	void removeElementIndexes(const DbXmlNodeImpl &node, Document &doc,
				  OperationContext &oc, bool reindex);
	void markElement(NidMap &map, const NsNid &, Document &doc,
			 bool attributes);
	void markElement(NidMap &map, const std::string &key, const NsNid &,
			 Document &doc, bool attributes);
	bool indexesRemoved(const std::string &key, bool attributes) const;
	void reindex(OperationContext &oc);
	void coalesceText(OperationContext &oc);
	
	void markForUpdate(Document *doc);
public:
	static NsNode *fetchNode(const DbXmlNodeImpl &node,
				 DbWrapper &db,
				 OperationContext &oc);
	static NsNode *fetchNode(const NsNid &nid,
				 const DocID &did,
				 DbWrapper &db,
				 OperationContext &oc);
private:
	void doCoalesceText(NsNode *node, Document &doc, OperationContext &oc);
	bool coalesceEntry(struct nsTextList *list, struct nsTextEntry *entries,
			   int &lastType, int index);
	static void putNode(const NsNode *node,
			    DbWrapper &db,
			    const DocID &did,
			    OperationContext &oc);
	static void deleteTree(NsNode *node,
			       DbWrapper &db,
			       const DocID &did,
			       OperationContext &oc);
	static void updateLastDescendants(NsDomNode *node,
					  const NsFullNid *nid,
					  DbWrapper &db,
					  const DocID &did,
					  OperationContext &oc);
private:
	UpdateMap attrMap_;
	UpdateMap textInsertMap_;
	UpdateMap textDeleteMap_;
	RewriteMap documentUpdates_;
	NidMap elements_;
	NidMap textCoalesce_;
};

}

#endif
