//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002,2008 Oracle.  All rights reserved.
//
// $Id: NodeVisitingOptimizer.cpp,v 1.50 2008/01/08 21:54:24 bostic Exp $
//

#include "DbXmlInternal.hpp"

#include "NodeVisitingOptimizer.hpp"
#include "../dataItem/Map.hpp"

#include "../query/ASTToQueryPlan.hpp"
#include "../query/QueryPlanToAST.hpp"
#include "../query/ParentJoinQP.hpp"
#include "../query/ChildJoinQP.hpp"
#include "../query/DescendantJoinQP.hpp"
#include "../query/AncestorJoinQP.hpp"
#include "../query/AttributeJoinQP.hpp"
#include "../query/IntersectQP.hpp"
#include "../query/PredicateFilterQP.hpp"
#include "../query/NodePredicateFilterQP.hpp"
#include "../query/UnionQP.hpp"
#include "../query/EmptyQP.hpp"
#include "../query/SequentialScanQP.hpp"
#include "../query/ValueFilterQP.hpp"
#include "../query/AttributeOrChildJoinQP.hpp"
#include "../query/ParentOfAttributeJoinQP.hpp"
#include "../query/ParentOfChildJoinQP.hpp"
#include "../query/DecisionPointQP.hpp"
#include "../query/ExceptQP.hpp"
#include "../query/ContextNodeQP.hpp"
#include "../query/CollectionQP.hpp"
#include "../query/DocQP.hpp"
#include "../query/BufferQP.hpp"
#include "../query/LevelFilterQP.hpp"
#include "../query/StepQP.hpp"
#include "../query/VariableQP.hpp"

using namespace DbXml;
using namespace std;

void NodeVisitingOptimizer::optimize(XQQuery *query)
{
	ASTVisitor::optimize(query);
}

ASTNode *NodeVisitingOptimizer::optimize(ASTNode *item)
{
	switch((DbXmlASTNode::whichType)item->getType()) {
	case (DbXmlASTNode::QP_TO_AST): {
		return optimizeQueryPlanToAST((QueryPlanToAST*)item);
		break;
	}
	case (DbXmlASTNode::NODE_CHECK): {
		return optimizeDbXmlNodeCheck((DbXmlNodeCheck*)item);
		break;
	}
	case (DbXmlASTNode::LAST_STEP_CHECK): {
		return optimizeDbXmlLastStepCheck((DbXmlLastStepCheck*)item);
		break;
	}
	case (DbXmlASTNode::MAP): {
		return optimizeMap((Map*)item);
		break;
	}
	case (DbXmlASTNode::DBXML_PREDICATE): {
		return optimizeDbXmlPredicate((DbXmlPredicate *)item);
		break;
	}
	}
	return ASTVisitor::optimize(item);
}

ASTNode *NodeVisitingOptimizer::optimizeQueryPlanToAST(QueryPlanToAST *item)
{
	item->setQueryPlan(optimizeQP(item->getQueryPlan()));
	return item;
}

ASTNode *NodeVisitingOptimizer::optimizeDbXmlNodeCheck(DbXmlNodeCheck *item)
{
	item->setArg(optimize(item->getArg()));
	return item;
}

ASTNode *NodeVisitingOptimizer::optimizeDbXmlLastStepCheck(DbXmlLastStepCheck *item)
{
	item->setArg(optimize(item->getArg()));
	return item;
}

ASTNode *NodeVisitingOptimizer::optimizeMap(Map *item)
{
	item->setArg1(optimize(item->getArg1()));
	item->setArg2(optimize(item->getArg2()));
	return item;
}

ASTNode *NodeVisitingOptimizer::optimizeDbXmlPredicate(DbXmlPredicate *item)
{
	item->setExpression(optimize(const_cast<ASTNode *>(item->getExpression())));
	item->setPredicate(optimize(const_cast<ASTNode *>(item->getPredicate())));
	return item;
}

////////////////////////////////////////////////////////////////////////////////////////////////////

QueryPlan *NodeVisitingOptimizer::optimizeQP(QueryPlan *item)
{
	QueryPlan *result = item;
	switch(item->getType()) {
	case QueryPlan::UNION: {
		result = optimizeUnion((UnionQP*)item);
		break;
	}
	case QueryPlan::INTERSECT: {
		result = optimizeIntersect((IntersectQP*)item);
		break;
	}
	case QueryPlan::PATHS: {
		result = optimizePaths((PathsQP*)item);
		break;
	}
	case QueryPlan::PRESENCE: {
		result = optimizePresence((PresenceQP*)item);
		break;
	}
	case QueryPlan::VALUE: {
		result = optimizeValue((ValueQP*)item);
		break;
	}
	case QueryPlan::RANGE: {
		result = optimizeRange((RangeQP*)item);
		break;
	}
	case QueryPlan::EMPTY: {
		result = optimizeEmpty((EmptyQP*)item);
		break;
	}
	case QueryPlan::SEQUENTIAL_SCAN: {
		result = optimizeSequentialScan((SequentialScanQP*)item);
		break;
	}
	case QueryPlan::VALUE_FILTER: {
		result = optimizeValueFilter((ValueFilterQP*)item);
		break;
	}
	case QueryPlan::PREDICATE_FILTER: {
		result = optimizePredicateFilter((PredicateFilterQP*)item);
		break;
	}
	case QueryPlan::NUMERIC_PREDICATE_FILTER: {
		result = optimizeNumericPredicateFilter((NumericPredicateFilterQP*)item);
		break;
	}
	case QueryPlan::NODE_PREDICATE_FILTER: {
		result = optimizeNodePredicateFilter((NodePredicateFilterQP*)item);
		break;
	}
	case QueryPlan::NEGATIVE_NODE_PREDICATE_FILTER: {
		result = optimizeNegativeNodePredicateFilter((NegativeNodePredicateFilterQP*)item);
		break;
	}
	case QueryPlan::AST_TO_QP: {
		result = optimizeASTToQueryPlan((ASTToQueryPlan*)item);
		break;
	}
	case QueryPlan::DESCENDANT: {
		result = optimizeDescendantJoin((DescendantJoinQP*)item);
		break;
	}
	case QueryPlan::DESCENDANT_OR_SELF: {
		result = optimizeDescendantOrSelfJoin((DescendantOrSelfJoinQP*)item);
		break;
	}
	case QueryPlan::ANCESTOR: {
		result = optimizeAncestorJoin((AncestorJoinQP*)item);
		break;
	}
	case QueryPlan::ANCESTOR_OR_SELF: {
		result = optimizeAncestorOrSelfJoin((AncestorOrSelfJoinQP*)item);
		break;
	}
	case QueryPlan::ATTRIBUTE: {
		result = optimizeAttributeJoin((AttributeJoinQP*)item);
		break;
	}
	case QueryPlan::CHILD: {
		result = optimizeChildJoin((ChildJoinQP*)item);
		break;
	}
	case QueryPlan::ATTRIBUTE_OR_CHILD: {
		result = optimizeAttributeOrChildJoin((AttributeOrChildJoinQP*)item);
		break;
	}
	case QueryPlan::PARENT: {
		result = optimizeParentJoin((ParentJoinQP*)item);
		break;
	}
	case QueryPlan::PARENT_OF_ATTRIBUTE: {
		result = optimizeParentOfAttributeJoin((ParentOfAttributeJoinQP*)item);
		break;
	}
	case QueryPlan::PARENT_OF_CHILD: {
		result = optimizeParentOfChildJoin((ParentOfChildJoinQP*)item);
		break;
	}
	case QueryPlan::DECISION_POINT: {
		result = optimizeDecisionPoint((DecisionPointQP*)item);
		break;
	}
	case QueryPlan::DECISION_POINT_END: {
		result = optimizeDecisionPointEnd((DecisionPointEndQP*)item);
		break;
	}
	case QueryPlan::EXCEPT: {
		result = optimizeExcept((ExceptQP*)item);
		break;
	}
	case QueryPlan::CONTEXT_NODE: {
		result = optimizeContextNode((ContextNodeQP*)item);
		break;
	}
	case QueryPlan::COLLECTION: {
		result = optimizeCollection((CollectionQP*)item);
		break;
	}
	case QueryPlan::DOC: {
		result = optimizeDoc((DocQP*)item);
		break;
	}
	case QueryPlan::BUFFER: {
		result = optimizeBuffer((BufferQP*)item);
		break;
	}
	case QueryPlan::BUFFER_REF: {
		result = optimizeBufferReference((BufferReferenceQP*)item);
		break;
	}
	case QueryPlan::LEVEL_FILTER: {
		result = optimizeLevelFilter((LevelFilterQP*)item);
		break;
	}
	case QueryPlan::STEP: {
		result = optimizeStepQP((StepQP*)item);
		break;
	}
	case QueryPlan::VARIABLE: {
		result = optimizeVariableQP((VariableQP*)item);
		break;
	}
	default: break;
	}
	return result;
}

QueryPlan *NodeVisitingOptimizer::optimizeUnion(UnionQP *item)
{
	OperationQP::Vector &args = const_cast<OperationQP::Vector&>(item->getArgs());
	for(OperationQP::Vector::iterator i = args.begin(); i != args.end(); ++i) {
		*i = optimizeQP(*i);
	}
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeIntersect(IntersectQP *item)
{
	OperationQP::Vector &args = const_cast<OperationQP::Vector&>(item->getArgs());
	for(OperationQP::Vector::iterator i = args.begin(); i != args.end(); ++i) {
		*i = optimizeQP(*i);
	}
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizePaths(PathsQP *item)
{
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizePresence(PresenceQP *item)
{
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeValue(ValueQP *item)
{
	QPValue &value = const_cast<QPValue&>(item->getValue());
	if(value.getASTNode() != 0) {
		value.setASTNode(optimize(const_cast<ASTNode*>(value.getASTNode())));
	}
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeRange(RangeQP *item)
{
	QPValue &value = const_cast<QPValue&>(item->getValue());
	if(value.getASTNode() != 0) {
		value.setASTNode(optimize(const_cast<ASTNode*>(value.getASTNode())));
	}
	QPValue &value2 = const_cast<QPValue&>(item->getValue2());
	if(value2.getASTNode() != 0) {
		value2.setASTNode(optimize(const_cast<ASTNode*>(value2.getASTNode())));
	}
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeEmpty(EmptyQP *item)
{
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeSequentialScan(SequentialScanQP *item)
{
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeValueFilter(ValueFilterQP *item)
{
	item->setArg(optimizeQP(item->getArg()));

	ImpliedSchemaNode *isn = item->getImpliedSchemaNode();
	isn->setASTNode(optimize(const_cast<ASTNode*>(isn->getASTNode())));

	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizePredicateFilter(PredicateFilterQP *item)
{
	item->setArg(optimizeQP(item->getArg()));
	item->setPred(optimize(item->getPred()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeNumericPredicateFilter(NumericPredicateFilterQP *item)
{
	item->setArg(optimizeQP(item->getArg()));
	item->setPred(optimize(item->getPred()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeNodePredicateFilter(NodePredicateFilterQP *item)
{
	item->setArg(optimizeQP(item->getArg()));
	item->setPred(optimizeQP(item->getPred()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeNegativeNodePredicateFilter(NegativeNodePredicateFilterQP *item)
{
	item->setArg(optimizeQP(item->getArg()));
	item->setPred(optimizeQP(item->getPred()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeASTToQueryPlan(ASTToQueryPlan *item)
{
	item->setASTNode(optimize(item->getASTNode()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeDescendantJoin(DescendantJoinQP *item)
{
	item->setLeftArg(optimizeQP(item->getLeftArg()));
	item->setRightArg(optimizeQP(item->getRightArg()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeDescendantOrSelfJoin(DescendantOrSelfJoinQP *item)
{
	item->setLeftArg(optimizeQP(item->getLeftArg()));
	item->setRightArg(optimizeQP(item->getRightArg()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeAncestorJoin(AncestorJoinQP *item)
{
	item->setLeftArg(optimizeQP(item->getLeftArg()));
	item->setRightArg(optimizeQP(item->getRightArg()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeAncestorOrSelfJoin(AncestorOrSelfJoinQP *item)
{
	item->setLeftArg(optimizeQP(item->getLeftArg()));
	item->setRightArg(optimizeQP(item->getRightArg()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeAttributeJoin(AttributeJoinQP *item)
{
	item->setLeftArg(optimizeQP(item->getLeftArg()));
	item->setRightArg(optimizeQP(item->getRightArg()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeChildJoin(ChildJoinQP *item)
{
	item->setLeftArg(optimizeQP(item->getLeftArg()));
	item->setRightArg(optimizeQP(item->getRightArg()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeAttributeOrChildJoin(AttributeOrChildJoinQP *item)
{
	item->setLeftArg(optimizeQP(item->getLeftArg()));
	item->setRightArg(optimizeQP(item->getRightArg()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeParentJoin(ParentJoinQP *item)
{
	item->setLeftArg(optimizeQP(item->getLeftArg()));
	item->setRightArg(optimizeQP(item->getRightArg()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeParentOfAttributeJoin(ParentOfAttributeJoinQP *item)
{
	item->setLeftArg(optimizeQP(item->getLeftArg()));
	item->setRightArg(optimizeQP(item->getRightArg()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeParentOfChildJoin(ParentOfChildJoinQP *item)
{
	item->setLeftArg(optimizeQP(item->getLeftArg()));
	item->setRightArg(optimizeQP(item->getRightArg()));
	return item;
}

DecisionPointSource *NodeVisitingOptimizer::optimizeDPS(DecisionPointSource *item)
{
	switch(item->getType()) {
	case DecisionPointSource::QUERY_PLAN:
		return optimizeQueryPlanDPSource((QueryPlanDPSource*)item);
	default: break;
	}
	return item;
}

DecisionPointSource *NodeVisitingOptimizer::optimizeQueryPlanDPSource(QueryPlanDPSource *item)
{
	item->setParent(optimizeQP(item->getParent()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeDecisionPoint(DecisionPointQP *item)
{
	if(item->getSource())
		item->setSource(optimizeDPS(item->getSource()));

	if(item->getArg() != 0)
		item->setArg(optimizeQP(item->getArg()));

	for(DecisionPointQP::ListItem *li = item->getList(); li != 0; li = li->next)
		li->qp = optimizeQP(li->qp);

	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeDecisionPointEnd(DecisionPointEndQP *item)
{
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeExcept(ExceptQP *item)
{
	item->setLeftArg(optimizeQP(item->getLeftArg()));
	item->setRightArg(optimizeQP(item->getRightArg()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeContextNode(ContextNodeQP *item)
{
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeCollection(CollectionQP *item)
{
	if(item->getArg() != 0)
		item->setArg(optimize(item->getArg()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeDoc(DocQP *item)
{
	item->setArg(optimize(item->getArg()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeBuffer(BufferQP *item)
{
	item->setParent(optimizeQP(item->getParent()));
	item->setArg(optimizeQP(item->getArg()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeBufferReference(BufferReferenceQP *item)
{
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeLevelFilter(LevelFilterQP *item)
{
	item->setArg(optimizeQP(item->getArg()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeStepQP(StepQP *item)
{
	item->setArg(optimizeQP(item->getArg()));
	return item;
}

QueryPlan *NodeVisitingOptimizer::optimizeVariableQP(VariableQP *item)
{
	return item;
}

