//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002,2008 Oracle.  All rights reserved.
//
// $Id: BufferQP.cpp,v 1.22 2008/01/08 21:54:25 bostic Exp $
//

#include "../DbXmlInternal.hpp"
#include "BufferQP.hpp"
#include "QueryPlanToAST.hpp"
#include "ASTToQueryPlan.hpp"
#include "ExceptQP.hpp"
#include "FilterQP.hpp"
#include "StructuralJoinQP.hpp"
#include "StepQP.hpp"
#include "QueryExecutionContext.hpp"
#include "../Container.hpp"
#include "../PrintDataItemTree.hpp"
#include "../dataItem/DbXmlConfiguration.hpp"
#include "../optimizer/NodeVisitingOptimizer.hpp"

#include <xqilla/context/DynamicContext.hpp>

#include <sstream>

using namespace DbXml;
using namespace std;

static const int INDENT = 1;

BufferQP::BufferQP(QueryPlan *parent, QueryPlan *arg, unsigned int bufferId, u_int32_t flags, XPath2MemoryManager *mm)
	: QueryPlan(BUFFER, flags, mm),
	  parent_(parent),
	  arg_(arg),
	  bufferId_(bufferId)
{
}

/** Sets the BufferQP object on all it's BufferReferenceQP objects */
class BufferReferenceSetter : public NodeVisitingOptimizer
{
public:
	virtual void run(BufferQP *b)
	{
		buf = b;
		NodeVisitingOptimizer::optimizeQP(buf->getArg());
	}

private:
	virtual QueryPlan *optimizeBufferReference(BufferReferenceQP *item)
	{
		if(item->getBufferID() == buf->getBufferID()) {
			item->setBuffer(buf);
		}
		return item;
	}

	virtual void resetInternal() {}

	BufferQP *buf;
};

QueryPlan *BufferQP::staticTyping(StaticContext *context)
{
	_src.clear();

	parent_ = parent_->staticTyping(context);
	// Do not add the StaticAnalysis, as
	// that is handled by the BufferEndQP objects

	BufferReferenceSetter().run(this);

	arg_ = arg_->staticTyping(context);
	_src.copy(arg_->getStaticAnalysis());

	return this;
}

void BufferQP::staticTypingLite(StaticContext *context)
{
	_src.clear();

	parent_->staticTypingLite(context);
	// Do not add the StaticAnalysis, as
	// that is handled by the BufferEndQP objects

	BufferReferenceSetter().run(this);

	arg_->staticTypingLite(context);
	_src.copy(arg_->getStaticAnalysis());
}

QueryPlan *BufferQP::optimize(OptimizationContext &opt)
{
	// Optimise the parent QueryPlan
	parent_ = parent_->optimize(opt);

	BufferReferenceSetter().run(this);

	arg_ = arg_->optimize(opt);

	return this;
}

static unsigned int countQP(QueryPlan *qp)
{
	switch(qp->getType()) {
	case QueryPlan::UNION:
	case QueryPlan::INTERSECT: {
		unsigned int result = 1;
		OperationQP::Vector &args = const_cast<OperationQP::Vector&>(((OperationQP*)qp)->getArgs());
		for(OperationQP::Vector::iterator i = args.begin(); i != args.end(); ++i) {
			result += countQP(*i);
		}
		return result;
	}
	case QueryPlan::EXCEPT:
		return 1 + countQP(((ExceptQP*)qp)->getLeftArg()) +
			countQP(((ExceptQP*)qp)->getRightArg());
	case QueryPlan::VALUE_FILTER:
	case QueryPlan::PREDICATE_FILTER:
	case QueryPlan::NODE_PREDICATE_FILTER:
	case QueryPlan::NEGATIVE_NODE_PREDICATE_FILTER:
	case QueryPlan::NUMERIC_PREDICATE_FILTER:
	case QueryPlan::LEVEL_FILTER:
		return 1 + countQP(((FilterQP*)qp)->getArg());
	case QueryPlan::STEP:
		return 1 + countQP(((StepQP*)qp)->getArg());
	case QueryPlan::DESCENDANT:
	case QueryPlan::DESCENDANT_OR_SELF:
	case QueryPlan::ANCESTOR:
	case QueryPlan::ANCESTOR_OR_SELF:
	case QueryPlan::ATTRIBUTE:
	case QueryPlan::CHILD:
	case QueryPlan::ATTRIBUTE_OR_CHILD:
	case QueryPlan::PARENT:
	case QueryPlan::PARENT_OF_ATTRIBUTE:
	case QueryPlan::PARENT_OF_CHILD:
		return 1 + countQP(((StructuralJoinQP*)qp)->getLeftArg()) +
			countQP(((StructuralJoinQP*)qp)->getRightArg());
	case QueryPlan::PRESENCE:
	case QueryPlan::VALUE:
	case QueryPlan::RANGE:
	case QueryPlan::SEQUENTIAL_SCAN:
	case QueryPlan::EMPTY:
	case QueryPlan::CONTEXT_NODE:
	case QueryPlan::VARIABLE:
	case QueryPlan::COLLECTION:
	case QueryPlan::DOC:
			return 1;
	case QueryPlan::AST_TO_QP:
	case QueryPlan::DECISION_POINT:
	case QueryPlan::DECISION_POINT_END:
	case QueryPlan::BUFFER:
	case QueryPlan::BUFFER_REF:
	case QueryPlan::PATHS: break;
	}
	return 1000000; // A very large number
}

#define INLINE_BYTES_PER_KEY_THRESHOLD 1024
#define INLINE_SIZE_THRESHOLD 20

class BufferUseCount : public NodeVisitingOptimizer
{
public:
	virtual int count(unsigned int id, QueryPlan *item)
	{
		id_ = id;
		count_ = 0;
		NodeVisitingOptimizer::optimizeQP(item);
		return count_;
	}

private:
	virtual QueryPlan *optimizeBufferReference(BufferReferenceQP *item)
	{
		if(item->getBufferID() == id_) ++count_;
		return item;
	}

	virtual void resetInternal() {}

	unsigned int id_;
	int count_;
};

class BufferRemover : public NodeVisitingOptimizer
{
public:
	virtual QueryPlan *run(const BufferQP *b, XPath2MemoryManager *m)
	{
		buf = b;
		mm = m;
		return NodeVisitingOptimizer::optimizeQP(buf->getArg()->copy(mm));
	}

private:
	virtual QueryPlan *optimizeBufferReference(BufferReferenceQP *item)
	{
		if(item->getBufferID() == buf->getBufferID()) {
			item->release();
			return buf->getParent()->copy(mm);
		}
		return item;
	}

	virtual void resetInternal() {}

	const BufferQP *buf;
	XPath2MemoryManager *mm;
};

void BufferQP::createCombinations(OptimizationContext &opt, QueryPlans &combinations) const
{
	XPath2MemoryManager *mm = opt.getMemoryManager();

	// This is deliberately performed on the original BufferQP object
	int useCount = BufferUseCount().count(bufferId_, arg_);
	if(useCount <= 1) {
		AutoRelease<QueryPlan> qp(BufferRemover().run(this, mm));
		qp->createCombinations(opt, combinations);
	} else {
		BufferQP *result = new (mm) BufferQP(parent_->chooseAlternative(opt, "buffer"), arg_->chooseAlternative(opt, "buffer"), bufferId_, flags_, mm);
		result->setLocationInfo(this);
		BufferReferenceSetter().run(result);
		combinations.push_back(result);
	}
}

void BufferQP::applyConversionRules(OptimizationContext &opt, QueryPlans &alternatives)
{
	XPath2MemoryManager *mm = opt.getMemoryManager();

	alternatives.push_back(this);

	int useCount = BufferUseCount().count(bufferId_, arg_);
	if(useCount <= 1 || countQP(parent_) < INLINE_SIZE_THRESHOLD) {
		AutoRelease<QueryPlan> qp(BufferRemover().run(this, mm));
		qp->createAlternatives(opt, alternatives);
	}
}

NodeIterator *BufferQP::createNodeIterator(DynamicContext *context) const
{
	return new BufferIterator(this, context);
}

QueryPlan *BufferQP::copy(XPath2MemoryManager *mm) const
{
	if(!mm) {
		mm = memMgr_;
	}

	BufferQP *result = new (mm) BufferQP(parent_->copy(mm), arg_->copy(mm), bufferId_, flags_, mm);
	result->setLocationInfo(this);
	BufferReferenceSetter().run(result);

	return result;
}

void BufferQP::release()
{
	parent_->release();
	arg_->release();
	memMgr_->deallocate(this);
}

void BufferQP::findQueryPlanRoots(QPRSet &qprset) const
{
	parent_->findQueryPlanRoots(qprset);
	arg_->findQueryPlanRoots(qprset);
}

Cost BufferQP::cost(OperationContext &context, QueryExecutionContext &qec) const
{
	Cost parentCost = parent_->cost(context, qec);

	Cost cost = arg_->cost(context, qec);
	cost.pagesOverhead += parentCost.totalPages();

	// Add the cost of storing the buffered data
	cost.pagesOverhead += parentCost.keys * INLINE_BYTES_PER_KEY_THRESHOLD;

	return cost;
}

bool BufferQP::isSubsetOf(const QueryPlan *o) const
{
	return arg_->isSubsetOf(o);
}

string BufferQP::printQueryPlan(const DynamicContext *context, int indent) const
{
	ostringstream s;

	string in(PrintAST::getIndent(indent));

	s << in << "<BufferQP id=\"" << bufferId_ << "\">" << endl;
	s << parent_->printQueryPlan(context, indent + INDENT);
	s << arg_->printQueryPlan(context, indent + INDENT);
	s << in << "</BufferQP>" << endl;

	return s.str();
}

string BufferQP::toString(bool brief) const
{
	ostringstream s;

	s << "BUF(";
	s << bufferId_ << ",";
	s << parent_->toString(brief) << ",";
	s << arg_->toString(brief);
	s << ")";

	return s.str();
}
 
////////////////////////////////////////////////////////////////////////////////////////////////////

NodeIterator *BufferReferenceQP::createNodeIterator(DynamicContext *context) const
{
	BufferSource *source = GET_CONFIGURATION(context)->getBufferSource();
	DBXML_ASSERT(source);
	return source->getBuffer(bufferId_);
}

QueryPlan *BufferReferenceQP::copy(XPath2MemoryManager *mm) const
{
	if(!mm) {
		mm = memMgr_;
	}

	BufferReferenceQP *result = new (mm) BufferReferenceQP(bufferId_, bqp_, flags_, mm);
	result->setLocationInfo(this);
	return result;
}

void BufferReferenceQP::release()
{
	memMgr_->deallocate(this);
}

QueryPlan *BufferReferenceQP::staticTyping(StaticContext *context)
{
	_src.clear();

	_src.copy(bqp_->getParent()->getStaticAnalysis());

	return this;
}

void BufferReferenceQP::staticTypingLite(StaticContext *context)
{
	_src.clear();
	_src.copy(bqp_->getParent()->getStaticAnalysis());
}

QueryPlan *BufferReferenceQP::optimize(OptimizationContext &opt)
{
	return this;
}

void BufferReferenceQP::findQueryPlanRoots(QPRSet &qprset) const
{
}

Cost BufferReferenceQP::cost(OperationContext &context, QueryExecutionContext &qec) const
{
	Cost cost;
	cost.keys = bqp_->getParent()->cost(context, qec).keys;

	// Add the cost of retriving the buffered data
	cost.pagesForKeys = cost.keys * INLINE_BYTES_PER_KEY_THRESHOLD;

	return cost;
}

bool BufferReferenceQP::isSubsetOf(const QueryPlan *o) const
{
	if(o->getType() == BUFFER_REF) {
		return bufferId_ == ((BufferReferenceQP*)o)->bufferId_;
	}

	return bqp_->getParent()->isSubsetOf(o);
}

string BufferReferenceQP::printQueryPlan(const DynamicContext *context, int indent) const
{
	ostringstream s;

	string in(PrintAST::getIndent(indent));

	s << in << "<BufferReferenceQP id=\"" << bufferId_ << "\"/>" << endl;

	return s.str();
}

string BufferReferenceQP::toString(bool brief) const
{
	ostringstream s;

	s << "BR(" << bufferId_ << ")";

	return s.str();
}
 
////////////////////////////////////////////////////////////////////////////////////////////////////

BufferIterator::BufferIterator(const BufferQP *qp, DynamicContext *context)
	: ProxyIterator(qp),
	  qp_(qp),
	  // TBD Make the use count actually be the number of times that the buffer will be accessed - jpcs
// 	  buf_(new QueryPlanToASTResult(qp->getParent()->createNodeIterator(context), qp), qp->getUseCount()),
	  buf_(new QueryPlanToASTResult(qp->getParent()->createNodeIterator(context), qp)),
	  parentSource_(GET_CONFIGURATION(context)->getBufferSource())
{
	// TBD implement this using a temporary BTree, rather than ResultBuffer - jpcs

	AutoBufferSourceReset reset(GET_CONFIGURATION(context), this);
	result_ = qp->getArg()->createNodeIterator(context);
}

BufferIterator::~BufferIterator()
{
	if (result_)
		delete result_;
}

NodeIterator *BufferIterator::getBuffer(unsigned id)
{
	if(id == qp_->getBufferID())
		return new ASTToQueryPlanIterator(buf_.createResult(), location_);

	DBXML_ASSERT(parentSource_);
	return parentSource_->getBuffer(id);
}

bool BufferIterator::next(DynamicContext *context)
{
	AutoBufferSourceReset reset(GET_CONFIGURATION(context), this);
	return result_->next(context);
}

bool BufferIterator::seek(int container, const DocID &did, const NsNid &nid, DynamicContext *context)
{
	AutoBufferSourceReset reset(GET_CONFIGURATION(context), this);
	return result_->seek(container, did, nid, context);
}
