//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002,2008 Oracle.  All rights reserved.
//
// $Id: DocQP.cpp,v 1.24 2008/02/20 17:20:19 gmf Exp $
//

#include "../DbXmlInternal.hpp"
#include "DocQP.hpp"
#include "EmptyQP.hpp"
#include "ASTToQueryPlan.hpp"
#include "QueryExecutionContext.hpp"
#include "../Container.hpp"
#include "../QueryContext.hpp"
#include "../OperationContext.hpp"
#include "../ReferenceMinder.hpp"
#include "../PrintDataItemTree.hpp"
#include "../UTF8.hpp"
#include "../dataItem/DbXmlUri.hpp"
#include "../dataItem/DbXmlConfiguration.hpp"
#include "../nodeStore/NsUtil.hpp"
#include "../optimizer/QueryPlanOptimizer.hpp"

#include <xqilla/context/DynamicContext.hpp>
#include <xqilla/exceptions/FunctionException.hpp>

#include <xercesc/util/XMLUri.hpp>

#include <sstream>

using namespace DbXml;
using namespace std;

XERCES_CPP_NAMESPACE_USE;

static const int INDENT = 1;

DocQP::DocQP(ASTNode *arg, ImpliedSchemaNode *isn, DynamicContext *context, u_int32_t flags, XPath2MemoryManager *mm)
	: QueryPlan(DOC, flags, mm),
	  QueryPlanRoot(0),
	  arg_(arg),
	  isn_(isn),
	  documentName_(0)
{
	DbXmlConfiguration *conf = GET_CONFIGURATION(context);

	if(arg_->isConstant()) {
		const XMLCh* currentUri = getUriArg(context);
		if(currentUri != 0) {
			DbXmlUri uri(context->getBaseURI(), currentUri, /*documentUri*/true);
			if(uri.isDbXmlScheme()) {
				if(uri.getDocumentName() == "") {
					ostringstream oss;
					oss << "The URI '" << uri.getResolvedUri() <<
						"' does not specifiy a document";
					throw XmlException(XmlException::DOCUMENT_NOT_FOUND, oss.str());
				}

				documentName_ = XMLString::replicate(uri.getDocumentName().c_str(), mm);

				try {
					XmlContainer containerWrapper =
						uri.openContainer(conf->getManager(),
								  conf->getTransaction());
					container_ = (Container*)containerWrapper;
					conf->getMinder()->addContainer((Container*)containerWrapper);

					if(container_->getContainer() != 0 && container_->getContainer()->isWholedocContainer()) {
						conf->addImpliedSchemaNode(container_->getContainerID(), documentName_, isn);
					}
				}
				catch(XmlException &e) {
					e.setLocationInfo(this);
					throw;
				}
			} else if(uri.isResolved()) {
				conf->addImpliedSchemaNode(uri.getResolvedUri(), isn);
			}
		}
	} else {
		// Turn document projection off everywhere - it's not safe to project
		// documents if we can't identify everywhere a document will be used at
		// compile time.
		conf->overrideProjection();
	}
}

DocQP::DocQP(ASTNode *arg, ImpliedSchemaNode *isn, ContainerBase *cont, const char *docName, u_int32_t flags, XPath2MemoryManager *mm)
	: QueryPlan(DOC, flags, mm),
	  QueryPlanRoot(cont),
	  arg_(arg),
	  isn_(isn),
	  documentName_(docName)
{
}

NodeIterator *DocQP::createNodeIterator(DynamicContext *context) const
{
	return new DocIterator(this);
}

QueryPlan *DocQP::copy(XPath2MemoryManager *mm) const
{
	if(!mm) {
		mm = memMgr_;
	}

	DocQP *result = new (mm) DocQP(arg_, isn_, container_, documentName_, flags_, mm);
	result->setLocationInfo(this);
	return result;
}

void DocQP::release()
{
	memMgr_->deallocate(this);
}

QueryPlan *DocQP::staticTyping(StaticContext *context)
{
	_src.clear();

	if(arg_ != 0) {
		arg_ = arg_->staticTyping(context);
		_src.add(arg_->getStaticAnalysis());
	}

	_src.getStaticType().flags = StaticType::DOCUMENT_TYPE;
	_src.availableDocumentsUsed(true);
	_src.setProperties(StaticAnalysis::DOCORDER | StaticAnalysis::GROUPED |
		StaticAnalysis::PEER | StaticAnalysis::SUBTREE | StaticAnalysis::ONENODE);

	return this;
}

void DocQP::staticTypingLite(StaticContext *context)
{
	_src.clear();

	if(arg_ != 0) {
		_src.add(arg_->getStaticAnalysis());
	}

	_src.getStaticType().flags = StaticType::DOCUMENT_TYPE;
	_src.availableDocumentsUsed(true);
	_src.setProperties(StaticAnalysis::DOCORDER | StaticAnalysis::GROUPED |
		StaticAnalysis::PEER | StaticAnalysis::SUBTREE | StaticAnalysis::ONENODE);
}

QueryPlan *DocQP::optimize(OptimizationContext &opt)
{
	XPath2MemoryManager *mm = opt.getMemoryManager();

	if(opt.getQueryPlanOptimizer() && arg_ != 0) {
		arg_ = opt.getQueryPlanOptimizer()->optimize(arg_);
	}

	QueryPlan *result;
	if(container_ == 0) {
		if(arg_->isConstant()) {
			result = new (mm) EmptyQP(0, mm);
		}
		return this;
	} else {
		result = new (mm) ValueQP(ImpliedSchemaNode::METADATA, 0, DbXml::metaDataName_uri_name, /*documentIndex*/false, Syntax::STRING,
			DbWrapper::EQUALITY, /*generalComp*/false, documentName_, NsUtil::nsStringLen((unsigned char*)documentName_), isn_, flags_, mm);
	}
	result->setLocationInfo(this);

	logTransformation(opt.getLog(), result);

	OptimizationContext newOpt(opt.getPhase(), opt.getContext(), opt.getQueryPlanOptimizer(), container_);
	return result->optimize(newOpt);
}

const XMLCh *DocQP::getUriArg(DynamicContext *context) const
{
	Item::Ptr arg = arg_->createResult(context)->next(context);
	if(arg.isNull()) return 0;

	// Get the uri argument
	const XMLCh *currentUri = arg->asString(context);

	if(currentUri == 0) {
		XQThrow(FunctionException,
			X("DocQP::getUriArg"),
			X("Error retrieving resource [err:FODC0002]"));
	}

	if(!XMLUri::isValidURI(true, currentUri)) {
		XQThrow(FunctionException,
			X("DocQP::getUriArg"),
			X("Invalid URI format [err:FODC0005]"));
	}

	return currentUri;
}

void DocQP::findQueryPlanRoots(QPRSet &qprset) const
{
	qprset.insert(this);
}

Cost DocQP::cost(OperationContext &context, QueryExecutionContext &qec) const
{
	// TBD we need some way to estimate this - jpcs
	return Cost(1, 1);
}

bool DocQP::isSubsetOf(const QueryPlan *o) const
{
	// TBD Can we do better here? - jpcs
	return false;
}

string DocQP::printQueryPlan(const DynamicContext *context, int indent) const
{
	ostringstream s;

	string in(PrintAST::getIndent(indent));

	s << in << "<DocQP>" << endl;
	s << DbXmlPrintAST::print(arg_, context, indent + INDENT);
	s << in << "</DocQP>" << endl;

	return s.str();
}

string DocQP::toString(bool brief) const
{
	ostringstream s;

	s << "DOC";

	return s.str();
}
 
////////////////////////////////////////////////////////////////////////////////////////////////////

DocIterator::~DocIterator()
{
	delete result_;
}

void DocIterator::init(DynamicContext *context)
{
	toDo_ = false;

	const XMLCh* currentUri = qp_->getUriArg(context);

	DbXmlUri uri(context->getBaseURI(), currentUri, /*documentUri*/true);
	if(uri.isDbXmlScheme()) {
		if(uri.getDocumentName() == "") {
			ostringstream oss;
			oss << "The URI '" << uri.getResolvedUri() <<
				"' does not specifiy a document";
			throw XmlException(XmlException::DOCUMENT_NOT_FOUND, oss.str());
		}

		DbXmlConfiguration *conf = GET_CONFIGURATION(context);
		try {
			cont_ = uri.openContainer(conf->getManager(),
						  conf->getTransaction());
		}
		catch(XmlException &e) {
			e.setLocationInfo(location_);
			throw;
		}

		// Perform a presence lookup on the built-in node-metadata-equality-string
		// index for the document name
		Container *container = (Container*)cont_;
		result_ = container->createDocumentIterator(
			context, location_, uri.getDocumentName().c_str(),
			uri.getDocumentName().length());
	} else {
		// If it's not a DB XML URI, use the default behaviour
		if(currentUri != 0)
			result_ = new ASTToQueryPlanIterator(context->resolveDocument(currentUri, location_), location_);
		else result_ = new EmptyIterator(location_);
	}
}

bool DocIterator::next(DynamicContext *context)
{
	if(toDo_) init(context);
	return result_->next(context);
}

bool DocIterator::seek(int container, const DocID &did, const NsNid &nid, DynamicContext *context)
{
	if(toDo_) init(context);
	return result_->seek(container, did, nid, context);
}

